/*
 * QEMU Block driver for  NBD
 *
 * Copyright (C) 2016 Red Hat, Inc.
 * Copyright (C) 2008 Bull S.A.S.
 *     Author: Laurent Vivier <Laurent.Vivier@bull.net>
 *
 * Some parts:
 *    Copyright (C) 2007 Anthony Liguori <anthony@codemonkey.ws>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "nbd-client.h"

#define HANDLE_TO_INDEX(bs, handle) ((handle) ^ ((uint64_t)(intptr_t)bs))
#define INDEX_TO_HANDLE(bs, index)  ((index)  ^ ((uint64_t)(intptr_t)bs))

static void nbd_recv_coroutines_enter_all(NBDClientSession *s)
{
    int i;

    for (i = 0; i < MAX_NBD_REQUESTS; i++) {
        if (s->recv_coroutine[i]) {
            qemu_coroutine_enter(s->recv_coroutine[i], NULL);
        }
    }
}

static void nbd_teardown_connection(BlockDriverState *bs)
{
    NBDClientSession *client = nbd_get_client_session(bs);

    if (!client->ioc) { /* Already closed */
        return;
    }

    /* finish any pending coroutines */
    qio_channel_shutdown(client->ioc,
                         QIO_CHANNEL_SHUTDOWN_BOTH,
                         NULL);
    nbd_recv_coroutines_enter_all(client);

    nbd_client_detach_aio_context(bs);
    object_unref(OBJECT(client->sioc));
    client->sioc = NULL;
    object_unref(OBJECT(client->ioc));
    client->ioc = NULL;
}

static void nbd_reply_ready(void *opaque)
{
    BlockDriverState *bs = opaque;
    NBDClientSession *s = nbd_get_client_session(bs);
    uint64_t i;
    int ret;

    if (!s->ioc) { /* Already closed */
        return;
    }

    if (s->reply.handle == 0) {
        /* No reply already in flight.  Fetch a header.  It is possible
         * that another thread has done the same thing in parallel, so
         * the socket is not readable anymore.
         */
        ret = nbd_receive_reply(s->ioc, &s->reply);
        if (ret == -EAGAIN) {
            return;
        }
        if (ret < 0) {
            s->reply.handle = 0;
            goto fail;
        }
    }

    /* There's no need for a mutex on the receive side, because the
     * handler acts as a synchronization point and ensures that only
     * one coroutine is called until the reply finishes.  */
    i = HANDLE_TO_INDEX(s, s->reply.handle);
    if (i >= MAX_NBD_REQUESTS) {
        goto fail;
    }

    if (s->recv_coroutine[i]) {
        qemu_coroutine_enter(s->recv_coroutine[i], NULL);
        return;
    }

fail:
    nbd_teardown_connection(bs);
}

static void nbd_restart_write(void *opaque)
{
    BlockDriverState *bs = opaque;

    qemu_coroutine_enter(nbd_get_client_session(bs)->send_coroutine, NULL);
}

static int nbd_co_send_request(BlockDriverState *bs,
                               NBDRequest *request,
                               QEMUIOVector *qiov, int offset)
{
    NBDClientSession *s = nbd_get_client_session(bs);
    AioContext *aio_context;
    int rc, ret, i;

    qemu_co_mutex_lock(&s->send_mutex);

    for (i = 0; i < MAX_NBD_REQUESTS; i++) {
        if (s->recv_coroutine[i] == NULL) {
            s->recv_coroutine[i] = qemu_coroutine_self();
            break;
        }
    }

    g_assert(qemu_in_coroutine());
    assert(i < MAX_NBD_REQUESTS);
    request->handle = INDEX_TO_HANDLE(s, i);

    if (!s->ioc) {
        qemu_co_mutex_unlock(&s->send_mutex);
        return -EPIPE;
    }

    s->send_coroutine = qemu_coroutine_self();
    aio_context = bdrv_get_aio_context(bs);

    aio_set_fd_handler(aio_context, s->sioc->fd, false,
                       nbd_reply_ready, nbd_restart_write, bs);
    if (qiov) {
        qio_channel_set_cork(s->ioc, true);
        rc = nbd_send_request(s->ioc, request);
        if (rc >= 0) {
            ret = nbd_wr_syncv(s->ioc, qiov->iov, qiov->niov,
                               offset, request->len, 0, false);
            if (ret != request->len) {
                rc = -EIO;
            }
        }
        qio_channel_set_cork(s->ioc, false);
    } else {
        rc = nbd_send_request(s->ioc, request);
    }
    aio_set_fd_handler(aio_context, s->sioc->fd, false,
                       nbd_reply_ready, NULL, bs);
    s->send_coroutine = NULL;
    qemu_co_mutex_unlock(&s->send_mutex);
    return rc;
}

static inline ssize_t read_sync(QIOChannel *ioc, void *buffer, size_t size)
{
    struct iovec iov = { .iov_base = buffer, .iov_len = size };
    /* Sockets are kept in blocking mode in the negotiation phase.  After
     * that, a non-readable socket simply means that another thread stole
     * our request/reply.  Synchronization is done with recv_coroutine, so
     * that this is coroutine-safe.
     */

    return nbd_wr_syncv(ioc, &iov, 1, 0, size, true, false);
}

static void nbd_co_receive_reply(NBDClientSession *s,
    NBDRequest *request, NBDReply *reply,
    QEMUIOVector *qiov, int offset,
    NBDExtent **pextents, unsigned *nb_extents)
{
    int ret;
    unsigned i;

    /* Wait until we're woken up by the read handler.  TODO: perhaps
     * peek at the next reply and avoid yielding if it's ours?  */
    qemu_coroutine_yield();
    *reply = s->reply;
    if (reply->handle != request->handle ||
        !s->ioc) {
        reply->simple = true;
        reply->error = EIO;
    } else {
        if (qiov) {
            if ((reply->simple ? reply->error == 0 :
                         reply->type == NBD_REPLY_TYPE_OFFSET_DATA)) {
                ret = nbd_wr_syncv(s->ioc, qiov->iov, qiov->niov,
                                   offset, request->len, 1, false);
                if (ret != request->len) {
                    reply->error = EIO;
                    return;
                }
            } else if (!reply->simple &&
                       reply->type == NBD_REPLY_TYPE_OFFSET_HOLE) {
                qemu_iovec_memset(qiov, 0, 0, request->len);
            }
        }

        if (reply->error == 0 && !reply->simple &&
            reply->type == NBD_REPLY_TYPE_BLOCK_STATUS &&
            pextents && nb_extents)
        {
            if (read_sync(s->ioc, &reply->context_id,
                          sizeof(reply->context_id)) !=
                sizeof(reply->context_id))
            {
                reply->error = EIO;
                return;
            }
            cpu_to_be32s(&reply->context_id);

            *nb_extents =
                (reply->length - sizeof(reply->context_id)) / sizeof(NBDExtent);
            *pextents = g_new(NBDExtent, *nb_extents);
            if (read_sync(s->ioc, *pextents, *nb_extents * sizeof(NBDExtent)) !=
                *nb_extents * sizeof(NBDExtent))
            {
                g_free(*pextents);
                *pextents = NULL;
                reply->error = -EIO;
                return;
            }

            for (i = 0; i < *nb_extents; ++i) {
                cpu_to_be32s(&((*pextents)[i].length));
                cpu_to_be32s(&((*pextents)[i].flags));
            }
        }

        /* Tell the read handler to read another header.  */
        s->reply.handle = 0;
    }
}

static void nbd_coroutine_start(NBDClientSession *s,
                                NBDRequest *request)
{
    /* Poor man semaphore.  The free_sema is locked when no other request
     * can be accepted, and unlocked after receiving one reply.  */
    if (s->in_flight >= MAX_NBD_REQUESTS - 1) {
        qemu_co_mutex_lock(&s->free_sema);
        assert(s->in_flight < MAX_NBD_REQUESTS);
    }
    s->in_flight++;

    /* s->recv_coroutine[i] is set as soon as we get the send_lock.  */
}

static void nbd_coroutine_end(NBDClientSession *s,
                              NBDRequest *request)
{
    int i = HANDLE_TO_INDEX(s, request->handle);
    s->recv_coroutine[i] = NULL;
    if (s->in_flight-- == MAX_NBD_REQUESTS) {
        qemu_co_mutex_unlock(&s->free_sema);
    }
}

static int nbd_co_readv_1(BlockDriverState *bs, int64_t sector_num,
                          int nb_sectors, QEMUIOVector *qiov,
                          int offset)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = {
        .type = NBD_CMD_READ,
        .flags = client->structured_reply ? NBD_CMD_FLAG_DF : 0,
    };
    NBDReply reply;
    ssize_t ret;

    request.from = sector_num * 512;
    request.len = nb_sectors * 512;

    nbd_coroutine_start(client, &request);
    ret = nbd_co_send_request(bs, &request, NULL, 0);
    if (ret < 0) {
        goto out;
    }

    nbd_co_receive_reply(client, &request, &reply, qiov, offset, NULL, NULL);
    if (reply.error != 0) {
        ret = -reply.error;
    }

    if (!reply.simple) {
        while (!(reply.flags & NBD_REPLY_FLAG_DONE)) {
            nbd_co_receive_reply(client, &request, &reply, qiov, offset,
                                 NULL, NULL);
            if (reply.error != 0) {
                ret = -reply.error;
            }
            if (reply.simple) {
                ret = -EIO;
                goto out;
            }
        }
    }

out:
    nbd_coroutine_end(client, &request);

    return ret;
}

static int nbd_co_writev_1(BlockDriverState *bs, int64_t sector_num,
                           int nb_sectors, QEMUIOVector *qiov,
                           int offset, int *flags)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = { .type = NBD_CMD_WRITE };
    NBDReply reply;
    ssize_t ret;

    if ((*flags & BDRV_REQ_FUA) && (client->nbdflags & NBD_FLAG_SEND_FUA)) {
        *flags &= ~BDRV_REQ_FUA;
        request.flags |= NBD_CMD_FLAG_FUA;
    }

    request.from = sector_num * 512;
    request.len = nb_sectors * 512;

    nbd_coroutine_start(client, &request);
    ret = nbd_co_send_request(bs, &request, qiov, offset);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, 0, NULL, NULL);
    }
    nbd_coroutine_end(client, &request);
    return -reply.error;
}

int nbd_client_co_readv(BlockDriverState *bs, int64_t sector_num,
                        int nb_sectors, QEMUIOVector *qiov)
{
    int offset = 0;
    int ret;
    while (nb_sectors > NBD_MAX_SECTORS) {
        ret = nbd_co_readv_1(bs, sector_num, NBD_MAX_SECTORS, qiov, offset);
        if (ret < 0) {
            return ret;
        }
        offset += NBD_MAX_SECTORS * 512;
        sector_num += NBD_MAX_SECTORS;
        nb_sectors -= NBD_MAX_SECTORS;
    }
    return nbd_co_readv_1(bs, sector_num, nb_sectors, qiov, offset);
}

int nbd_client_co_writev(BlockDriverState *bs, int64_t sector_num,
                         int nb_sectors, QEMUIOVector *qiov, int *flags)
{
    int offset = 0;
    int ret;
    while (nb_sectors > NBD_MAX_SECTORS) {
        ret = nbd_co_writev_1(bs, sector_num, NBD_MAX_SECTORS, qiov, offset,
                              flags);
        if (ret < 0) {
            return ret;
        }
        offset += NBD_MAX_SECTORS * 512;
        sector_num += NBD_MAX_SECTORS;
        nb_sectors -= NBD_MAX_SECTORS;
    }
    return nbd_co_writev_1(bs, sector_num, nb_sectors, qiov, offset, flags);
}

int nbd_client_co_flush(BlockDriverState *bs)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = { .type = NBD_CMD_FLUSH };
    NBDReply reply;
    ssize_t ret;

    if (!(client->nbdflags & NBD_FLAG_SEND_FLUSH)) {
        return 0;
    }

    request.from = 0;
    request.len = 0;

    nbd_coroutine_start(client, &request);
    ret = nbd_co_send_request(bs, &request, NULL, 0);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, 0, NULL, NULL);
    }
    nbd_coroutine_end(client, &request);
    return -reply.error;
}

int nbd_client_co_discard(BlockDriverState *bs, int64_t sector_num,
                          int nb_sectors)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = { .type = NBD_CMD_TRIM };
    NBDReply reply;
    ssize_t ret;

    if (!(client->nbdflags & NBD_FLAG_SEND_TRIM)) {
        return 0;
    }
    request.from = sector_num * 512;
    request.len = nb_sectors * 512;

    nbd_coroutine_start(client, &request);
    ret = nbd_co_send_request(bs, &request, NULL, 0);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, 0, NULL, NULL);
    }
    nbd_coroutine_end(client, &request);
    return -reply.error;
}

int nbd_client_co_write_zeroes(BlockDriverState *bs, int64_t sector_num,
                          int nb_sectors, BdrvRequestFlags flags)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = { .type = NBD_CMD_WRITE_ZEROES };
    NBDReply reply;
    ssize_t ret;

    if (!(client->nbdflags & NBD_FLAG_SEND_WRITE_ZEROES)) {
        return -ENOTSUP;
    }

    if (flags & BDRV_REQ_FUA) {
        assert(client->nbdflags & NBD_FLAG_SEND_FUA);
        request.flags |= NBD_CMD_FLAG_FUA;
    }
    if (!(flags & BDRV_REQ_MAY_UNMAP)) {
        request.flags |= NBD_CMD_FLAG_NO_HOLE;
    }

    request.from = sector_num * 512;
    request.len = nb_sectors * 512;

    nbd_coroutine_start(client, &request);
    ret = nbd_co_send_request(bs, &request, NULL, 0);
    if (ret < 0) {
        reply.error = -ret;
    } else {
        nbd_co_receive_reply(client, &request, &reply, NULL, 0, NULL, NULL);
    }
    nbd_coroutine_end(client, &request);
    return -reply.error;
}

static int nbd_client_co_cmd_block_status(BlockDriverState *bs, uint64_t offset,
                                          uint64_t bytes, NBDExtent **pextents,
                                          unsigned *nb_extents)
{
    int64_t ret;
    NBDReply reply;
    unsigned nb;
    NBDExtent *extents = NULL;
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = {
        .type = NBD_CMD_BLOCK_STATUS,
        .from = offset,
        .len = bytes,
        .flags = 0,
    };

    nbd_coroutine_start(client, &request);

    ret = nbd_co_send_request(bs, &request, NULL, 0);
    if (ret < 0) {
        goto fail;
    }

    nbd_co_receive_reply(client, &request, &reply, NULL, 0, &extents, &nb);
    if (reply.error != 0) {
        ret = -reply.error;
        goto fail;
    }
    if (reply.simple) {
        ret = -EINVAL;
        goto fail;
    }
    if (reply.type != NBD_REPLY_TYPE_BLOCK_STATUS) {
        ret = -EINVAL;
        goto fail;
    }

    if (client->meta_data_context_id != reply.context_id) {
        ret = -EINVAL;
        goto fail;
    }

    if (!(reply.flags & NBD_REPLY_FLAG_DONE)) {
        nbd_co_receive_reply(client, &request, &reply, NULL, 0, NULL, NULL);
        if (reply.simple) {
            ret = -EINVAL;
            goto fail;
        }
        if (reply.error != 0) {
            ret = -reply.error;
            goto fail;
        }
        if (reply.type != NBD_REPLY_TYPE_NONE ||
            !(reply.flags & NBD_REPLY_FLAG_DONE)) {
            ret = -EINVAL;
            goto fail;
        }
    }

    *pextents = extents;
    *nb_extents = nb;
    nbd_coroutine_end(client, &request);
    return 0;

fail:
    g_free(extents);
    nbd_coroutine_end(client, &request);
    return ret;
}

/* nbd_client_co_load_bitmap_part() returns end of set area, i.e. first next
 * byte of unknown status (may be >= disk size, which means that the bitmap was
 * set up to the end).
 */
int64_t nbd_client_co_load_bitmap_part(BlockDriverState *bs, uint64_t offset,
                                       uint64_t bytes, BdrvDirtyBitmap *bitmap)
{
    int64_t ret;
    uint64_t start_byte;
    uint32_t nb_extents;
    int64_t i, start_sector, last_sector, nr_sectors;
    NBDExtent *extents = NULL;

    ret = nbd_client_co_cmd_block_status(bs, offset, bytes, &extents,
                                         &nb_extents);
    if (ret < 0) {
        return ret;
    }

    start_byte = offset;
    for (i = 0; i < nb_extents; ++i) {
        if (extents[i].flags == 1) {
            start_sector = start_byte >> BDRV_SECTOR_BITS;
            last_sector =
                (start_byte + extents[i].length - 1) >> BDRV_SECTOR_BITS;
            nr_sectors = last_sector - start_sector + 1;

            bdrv_set_dirty_bitmap(bitmap, start_sector, nr_sectors);
        }

        start_byte += extents[i].length;
    }

    g_free(extents);

    return ROUND_UP((uint64_t)start_byte,
                    (uint64_t)bdrv_dirty_bitmap_granularity(bitmap));
}

int64_t coroutine_fn nbd_client_co_get_block_status(BlockDriverState *bs,
                                                    int64_t sector_num,
                                                    int nb_sectors, int *pnum,
                                                    BlockDriverState **file)
{
    int64_t ret;
    uint32_t nb_extents;
    NBDExtent *extents;
    NBDClientSession *client = nbd_get_client_session(bs);

    if (!client->block_status_ok) {
        *pnum = nb_sectors;
        ret = BDRV_BLOCK_DATA | BDRV_BLOCK_ALLOCATED;
        if (bs->drv->protocol_name) {
            ret |= BDRV_BLOCK_OFFSET_VALID | (sector_num * BDRV_SECTOR_SIZE);
        }
        return ret;
    }

    ret = nbd_client_co_cmd_block_status(bs, sector_num << BDRV_SECTOR_BITS,
                                         nb_sectors << BDRV_SECTOR_BITS,
                                         &extents, &nb_extents);
    if (ret < 0) {
        return ret;
    }

    *pnum = extents[0].length >> BDRV_SECTOR_BITS;
    ret = (extents[0].flags & NBD_STATE_HOLE ? 0 : BDRV_BLOCK_ALLOCATED) |
          (extents[0].flags & NBD_STATE_ZERO ? BDRV_BLOCK_ZERO : 0);

    if ((ret & BDRV_BLOCK_ALLOCATED) && !(ret & BDRV_BLOCK_ZERO)) {
        ret |= BDRV_BLOCK_DATA;
    }

    g_free(extents);

    return ret;
}

void nbd_client_detach_aio_context(BlockDriverState *bs)
{
    aio_set_fd_handler(bdrv_get_aio_context(bs),
                       nbd_get_client_session(bs)->sioc->fd,
                       false, NULL, NULL, NULL);
}

void nbd_client_attach_aio_context(BlockDriverState *bs,
                                   AioContext *new_context)
{
    aio_set_fd_handler(new_context, nbd_get_client_session(bs)->sioc->fd,
                       false, nbd_reply_ready, NULL, bs);
}

void nbd_client_close(BlockDriverState *bs)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    NBDRequest request = { .type = NBD_CMD_DISC };

    if (client->ioc == NULL) {
        return;
    }

    nbd_send_request(client->ioc, &request);

    nbd_teardown_connection(bs);
}

int nbd_client_init(BlockDriverState *bs,
                    QIOChannelSocket *sioc,
                    const char *export,
                    QCryptoTLSCreds *tlscreds,
                    const char *hostname,
                    const char *bitmap_name,
                    Error **errp)
{
    NBDClientSession *client = nbd_get_client_session(bs);
    int ret;

    /* NBD handshake */
    logout("session init %s\n", export);
    qio_channel_set_blocking(QIO_CHANNEL(sioc), true, NULL);

    ret = nbd_receive_negotiate(QIO_CHANNEL(sioc), export,
                                &client->nbdflags,
                                tlscreds, hostname,
                                &client->ioc,
                                &client->size,
                                &client->structured_reply,
                                bitmap_name,
                                &client->bitmap_ok,
                                &client->block_status_ok, errp);
    if (ret < 0) {
        logout("Failed to negotiate with the NBD server\n");
        return ret;
    }

    qemu_co_mutex_init(&client->send_mutex);
    qemu_co_mutex_init(&client->free_sema);
    client->sioc = sioc;
    object_ref(OBJECT(client->sioc));

    if (!client->ioc) {
        client->ioc = QIO_CHANNEL(sioc);
        object_ref(OBJECT(client->ioc));
    }

    /* Now that we're connected, set the socket to be non-blocking and
     * kick the reply mechanism.  */
    qio_channel_set_blocking(QIO_CHANNEL(sioc), false, NULL);

    nbd_client_attach_aio_context(bs, bdrv_get_aio_context(bs));

    logout("Established connection with NBD server\n");
    return 0;
}
