/*
 * QEMU Hyper-V heartbeat device support
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 */

#include "qemu/osdep.h"
#include "qemu/error-report.h"
#include "qapi/error.h"
#include "hw/vmbus/vmbus.h"
#include "standard-headers/asm-x86/hyperv.h"

#define HV_HEARTBEAT_CLASS_UUID "57164f39-9115-4e78-ab55-382f3bd5422d"
#define HV_HEARTBEAT_INSTANCE_UUID "fd149e91-82e0-4a7d-afa6-2a4166cbd7c0"

#define HEARTBEAT_TICK_MS 2000

typedef struct VMBusHeartbeat
{
    VMBusDevice parent;

    QEMUTimer *timer;
    bool version_negotiated;
    uint16_t icsvc_ver_major;
    uint16_t icsvc_ver_minor;
    uint16_t hb_ver_major;
    uint16_t hb_ver_minor;
    uint64_t seq_num;
    bool alive;
} VMBusHeartbeat;

struct icmsg_packet {
    struct vmbuspipe_hdr pipe_hdr;
    struct icmsg_hdr icmsg_hdr;
    uint8_t msg_data[0];
};

/* ICMSG packets use global transaction id which is incremented on host.
 * As far as we know it is used only for debugging traffic dumps */
static uint32_t g_transaction_id;

static uint32_t icmsg_packet_size(struct icmsg_packet *packet)
{
        return sizeof(*packet) + packet->icmsg_hdr.icmsgsize;
}

static struct icmsg_packet *icmsg_packet_create(uint16_t msg_type,
        uint16_t msg_data_size,
        uint8_t flags)
{
    struct icmsg_packet *packet;

    packet = g_malloc0(sizeof(*packet) + msg_data_size);
    packet->pipe_hdr.flags = 1;
    packet->pipe_hdr.msgsize = sizeof(packet->icmsg_hdr) + msg_data_size;
    packet->icmsg_hdr.icmsgtype = msg_type;
    packet->icmsg_hdr.icmsgsize = msg_data_size;
    packet->icmsg_hdr.icflags = flags;
    packet->icmsg_hdr.ictransaction_id = atomic_fetch_inc(&g_transaction_id);

    return packet;
}

/* Handle negotiate version reply */
static void heartbeat_handle_version(VMBusHeartbeat *hbdev,
        const struct icmsg_packet *packet)
{
    assert(hbdev);
    assert(packet);

    struct icmsg_negotiate *neg_msg = (struct icmsg_negotiate*)packet->msg_data;

    if (hbdev->version_negotiated) {
        error_report("invalid device state - version already negotiated");
        return;
    }

    /* Guest should pick a single version for icframe and icmsg
     * In any case use the first one */
    if (neg_msg->icframe_vercnt < 1 || neg_msg->icmsg_vercnt < 1) {
        VMBUS_TRACE("guest did not pick a version");
        return;
    }

    hbdev->icsvc_ver_major = neg_msg->icversion_data[0].major;
    hbdev->icsvc_ver_minor = neg_msg->icversion_data[0].minor;
    hbdev->hb_ver_major = neg_msg->icversion_data[neg_msg->icframe_vercnt].major;
    hbdev->hb_ver_minor = neg_msg->icversion_data[neg_msg->icframe_vercnt].minor;
    hbdev->version_negotiated = true;

    VMBUS_TRACE("negotiated versions: icframe = %d.%d, icmsg = %d.%d",
            hbdev->icsvc_ver_major, hbdev->icsvc_ver_minor,
            hbdev->hb_ver_major, hbdev->hb_ver_minor);
}

/* Handle tick reply */
static void heartbeat_handle_tick(VMBusHeartbeat *hbdev,
        const struct icmsg_packet *packet)
{
    assert(hbdev);
    assert(packet);

    struct heartbeat_msg_data *hb_msg =
        (struct heartbeat_msg_data *) packet->msg_data;

    hbdev->alive = (hb_msg->seq_num == (hbdev->seq_num + 1));
    hbdev->seq_num = hb_msg->seq_num;

    VMBUS_TRACE("heartbeat chan %u alive %u seq_num %"PRIu64,
            hbdev->parent.chan->relid, hbdev->alive, hbdev->seq_num);
}

static void heartbeat_recv(VMBusHeartbeat *hbdev)
{
    int res = 0;

    assert(hbdev);

    uint32_t msglen = 0;
    res = vmbus_channel_recv_data(hbdev->parent.chan, NULL, 0, &msglen);
    if (res != -ENOSPC) {
        return;
    }

    VMBUS_TRACE("receiving message len %d", msglen);

    struct icmsg_packet *packet = g_malloc(msglen);
    res = vmbus_channel_recv_data(hbdev->parent.chan, packet, msglen, &msglen);
    if (res != 0) {
        error_report("channel recv failed with %d", res);
        goto out;
    }

    switch (packet->icmsg_hdr.icmsgtype) {
        case ICMSGTYPE_NEGOTIATE:
            heartbeat_handle_version(hbdev, packet);
            break;

        case ICMSGTYPE_HEARTBEAT:
            heartbeat_handle_tick(hbdev, packet);
            break;

        default:
            error_report("Unknown heartbeat response type %d",
                    packet->icmsg_hdr.icmsgtype);
            break;
    }

out:
    g_free(packet);
}

/* Guest signals available reply packet */
static void heartbeat_signal_event(EventNotifier *e)
{
    if (event_notifier_test_and_clear(e)) {
        VMBusDevice *dev = container_of(e, VMBusDevice, signal_event);
        VMBUS_TRACE("%p", dev);
        heartbeat_recv(VMBUS_HEARTBEAT(dev));
    }
}

static void heartbeat_negotiate_version(VMBusHeartbeat *hbdev)
{
    int err;
    struct icmsg_packet *packet;
    struct icmsg_negotiate *neg_msg;

    /*
     * Those versions have no real differences and guest will usually
     * pick the one that corresponds to its OS version.
     *
     * Negotiation packet will have two sets of versions: framework and device
     * Both sets advertise same version numbers.
     */
    const uint32_t versions[] = {
        VMBUS_VERSION_WS2008,
        VMBUS_VERSION_WIN7,
        VMBUS_VERSION_WIN8,
        VMBUS_VERSION_WIN8_1,
        VMBUS_VERSION_WIN10,
    };

    size_t vercnt = ARRAY_SIZE(versions);

    packet = icmsg_packet_create(ICMSGTYPE_NEGOTIATE,
            sizeof(*neg_msg) + vercnt * sizeof(struct ic_version) * 2,
            ICMSGHDRFLAG_TRANSACTION | ICMSGHDRFLAG_REQUEST);

    neg_msg = (struct icmsg_negotiate *)packet->msg_data;
    neg_msg->icframe_vercnt = vercnt;
    neg_msg->icmsg_vercnt = vercnt;

    int i;
    for (i = 0; i < vercnt; ++i) {
        neg_msg->icversion_data[i].major = (uint16_t)(versions[i] >> 16);
        neg_msg->icversion_data[i].minor = (uint16_t)versions[i];
        neg_msg->icversion_data[vercnt + i] = neg_msg->icversion_data[i];
    }

    err = vmbus_channel_send(hbdev->parent.chan, packet,
            icmsg_packet_size(packet), VM_PKT_DATA_INBAND, g_transaction_id);
    if (err) {
        error_report("channel send failed with %d", err);
    }
}

static void heartbeat_send_tick(VMBusHeartbeat *hbdev)
{
    int err;
    struct icmsg_packet *packet;
    struct heartbeat_msg_data *hb_msg;

    packet = icmsg_packet_create(ICMSGTYPE_HEARTBEAT, sizeof(*hb_msg),
            ICMSGHDRFLAG_REQUEST);
    packet->icmsg_hdr.icverframe.major = hbdev->icsvc_ver_major;
    packet->icmsg_hdr.icverframe.minor = hbdev->icsvc_ver_minor;
    packet->icmsg_hdr.icvermsg.major = hbdev->hb_ver_major;
    packet->icmsg_hdr.icvermsg.minor = hbdev->hb_ver_minor;

    hb_msg = (struct heartbeat_msg_data *)packet->msg_data;

    if (hbdev->seq_num == 0) {
        hbdev->seq_num = 2;
    } else if (hbdev->seq_num & 0x1) {
        hbdev->seq_num++;
    }
    hb_msg->seq_num = hbdev->seq_num;

    err = vmbus_channel_send(hbdev->parent.chan, packet,
            icmsg_packet_size(packet), VM_PKT_DATA_INBAND, g_transaction_id);

    if (err) {
        error_report("channel send failed with %d", err);
    }
}

/* Timer callback, every 2 seconds */
static void heartbeat_timer_fn(void* opaque)
{
    VMBusHeartbeat *hbdev = (VMBusHeartbeat*)opaque;

    if (!hbdev->version_negotiated) {
        /* Spam host with version negotiation packets until it replies */
        heartbeat_negotiate_version(hbdev);
    } else {
        heartbeat_send_tick(hbdev);
    }

    timer_mod(hbdev->timer,
            qemu_clock_get_ms(QEMU_CLOCK_HOST) + HEARTBEAT_TICK_MS);
}

static void heartbeat_reset(VMBusHeartbeat *hbdev)
{
    if (hbdev) {
        if (hbdev->timer) {
            timer_del(hbdev->timer);
        }

        hbdev->version_negotiated = false;
        hbdev->alive = false;
        hbdev->seq_num = 0;
    }
}

static int heartbeat_open_channel(VMBusDevice *dev, unsigned char *userdata)
{
    VMBUS_TRACE("%p", dev);
    VMBusHeartbeat *hb = VMBUS_HEARTBEAT(dev);
    timer_mod(hb->timer, qemu_clock_get_ms(0));
    return 0;
}

static void heartbeat_close_channel(VMBusDevice *dev)
{
    VMBUS_TRACE("dev %p", dev);
    VMBusHeartbeat *hb = VMBUS_HEARTBEAT(dev);
    heartbeat_reset(hb);
}

///////////////////////////////////////////////////////////////////////////////

static const VMStateDescription vmbus_heartbeat_state_description = {
    .name = TYPE_VMBUS_HEARTBEAT,
    .version_id = 1,
    .minimum_version_id = 1,
    .fields = (VMStateField[]) {
        VMSTATE_TIMER_PTR(timer, VMBusHeartbeat),
        VMSTATE_BOOL(version_negotiated, VMBusHeartbeat),
        VMSTATE_UINT16(icsvc_ver_major, VMBusHeartbeat),
        VMSTATE_UINT16(icsvc_ver_minor, VMBusHeartbeat),
        VMSTATE_UINT16(hb_ver_major, VMBusHeartbeat),
        VMSTATE_UINT16(hb_ver_minor, VMBusHeartbeat),
        VMSTATE_UINT64(seq_num, VMBusHeartbeat),
        VMSTATE_BOOL(alive, VMBusHeartbeat),
        VMSTATE_END_OF_LIST()
    }
};

static void vmbus_heartbeat_realize(DeviceState *dev, Error **errp)
{
    int res = 0;
    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusHeartbeat *hb = VMBUS_HEARTBEAT(vdev);
    Error *local_err = NULL;

    VMBUS_TRACE("dev %p", dev);

    res = qemu_uuid_parse(HV_HEARTBEAT_INSTANCE_UUID, &vdev->instanceid);
    assert(res == 0); /* Fixed UUID should definitely be correct */

    uint16_t chn_flags = VMBUS_CHANNEL_ENUMERATE_DEVICE_INTERFACE
            | VMBUS_CHANNEL_NAMED_PIPE_MODE;

    vdev->chan = vmbus_create_channel(vdev, chn_flags);
    if (!vdev->chan) {
        error_setg(&local_err, "failed to create channel");
        goto error_out;
    }

    res = event_notifier_init(&vdev->signal_event, 0);
    if (res != 0) {
        error_setg(&local_err, "event notifier failed to init with %d", res);
        goto error_out;
    }

    res = event_notifier_set_handler(&vdev->signal_event, false, heartbeat_signal_event);
    if (res != 0) {
        error_setg(&local_err, "event notifier set handler failed with %d", res);
        goto error_out;
    }

    res = vmstate_register(dev, -1, &vmbus_heartbeat_state_description, dev);
    if (res != 0) {
        error_setg(&local_err, "vmstate_register says %d", res);
        goto error_out;
    }

    hb->timer = timer_new_ms(QEMU_CLOCK_HOST, heartbeat_timer_fn, hb);
    return;

error_out:
    vmbus_release_channel(vdev->chan);
    event_notifier_cleanup(&vdev->signal_event);
    error_propagate(errp, local_err);
}

static void vmbus_heartbeat_reset(DeviceState *dev)
{
    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusHeartbeat *hb = VMBUS_HEARTBEAT(vdev);
    VMBUS_TRACE("%p", dev);

    heartbeat_reset(hb);
}

static void vmbus_heartbeat_unrealize(DeviceState *dev, Error **errp)
{
    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusHeartbeat *hb = VMBUS_HEARTBEAT(vdev);
    VMBUS_TRACE("%p", dev);

    timer_free(hb->timer);
}

static void vmbus_heartbeat_class_init(ObjectClass *klass, void *data)
{
    VMBusDeviceClass *vdc = VMBUS_DEVICE_CLASS(klass);

    VMBUS_TRACE("");

    int res = qemu_uuid_parse(HV_HEARTBEAT_CLASS_UUID, &vdc->class_uuid);
    assert(res == 0);
    vdc->singleton = true;
    vdc->realize = vmbus_heartbeat_realize;
    vdc->unrealize = vmbus_heartbeat_unrealize;
    vdc->reset = vmbus_heartbeat_reset;
    vdc->open_channel = heartbeat_open_channel;
    vdc->close_channel = heartbeat_close_channel;
}

static const TypeInfo vmbus_heartbeat_type_info = {
    .name = TYPE_VMBUS_HEARTBEAT,
    .parent = TYPE_VMBUS_DEVICE,
    .instance_size = sizeof(VMBusHeartbeat),
    .class_init = vmbus_heartbeat_class_init,
};

static void vmbus_heartbeat_register_types(void)
{
    type_register_static(&vmbus_heartbeat_type_info);
}

type_init(vmbus_heartbeat_register_types)
