/*
 * QEMU Hyper-V VMBus support
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 */

#include <qemu/osdep.h>
#include <target-i386/cpu.h>
#include "hw/vmbus/pagemap.h"
#include "hw/vmbus/vmbus.h"

void *hv_page_map_map(hv_page_map *map, uint32_t pg_index, bool rw)
{
    assert(map);

    if (pg_index >= map->nr_pages) {
        return NULL;
    }

    hv_page_map_slot *slot = &map->pg_slot[pg_index];
    if (slot->address) {
        return slot->address;
    }

    uint64_t mapped_len = TARGET_PAGE_SIZE;
    void *page = cpu_physical_memory_map(slot->pfn * TARGET_PAGE_SIZE,
            &mapped_len, rw);
    if (!page) {
        return NULL;
    } else if (mapped_len != TARGET_PAGE_SIZE) {
        cpu_physical_memory_unmap(page, mapped_len, rw, 0);
        return NULL;
    }

    VMBUS_TRACE("mapped pfn %#" PRIx64 " at %p", slot->pfn, page);
    slot->address = page;
    slot->rw = rw;
    return page;
}

void hv_page_map_unmap(hv_page_map *map, uint32_t pg_index)
{
    assert(map);

    if (pg_index >= map->nr_pages) {
        return;
    }

    hv_page_map_slot *slot = &map->pg_slot[pg_index];
    if (!slot->address) {
        return;
    }

    cpu_physical_memory_unmap(slot->address, TARGET_PAGE_SIZE, slot->rw,
            TARGET_PAGE_SIZE);
    VMBUS_TRACE("unmapped pfn %#" PRIx64 " at %p", slot->pfn, slot->address);
    slot->address = NULL;
}

static bool page_map_rw(hv_page_map *map, uint32_t pos,
        void *buf, uint32_t len, bool is_read)
{
    assert(map);
    assert(buf);

    uint32_t page = pos / TARGET_PAGE_SIZE;
    uint32_t offset = pos & (TARGET_PAGE_SIZE - 1);
    uint8_t *pbuf = buf;

    while (len > 0) {
        uint8_t* data = hv_page_map_map(map, page, !is_read);
        if (!data) {
            return false;
        }

        uint32_t nbytes = (TARGET_PAGE_SIZE - offset > len ?
                len : TARGET_PAGE_SIZE - offset);

        if (is_read) {
            memcpy(pbuf, data + offset, nbytes);
        } else {
            memcpy(data + offset, pbuf, nbytes);
        }

        /* Do not unmap page here if we've mapped it.
         * Combine all unmapping on cleanup instead */

        page++;
        pbuf += nbytes;
        len -= nbytes;
        offset = 0;
    }

    return true;
}

bool hv_page_map_read(hv_page_map *map, uint32_t pos,
        void *dst, uint32_t len)
{
    return page_map_rw(map, pos, dst, len, true);
}

bool hv_page_map_write(hv_page_map *map, uint32_t pos,
        void *src, uint32_t len)
{
    return page_map_rw(map, pos, src, len, false);
}

static hv_page_map *alloc_map(uint32_t nr_pages)
{
    hv_page_map *map = g_malloc0(sizeof(*map) + nr_pages * sizeof(map->pg_slot[0]));
    map->nr_pages = nr_pages;
    return map;
}

hv_page_map *hv_page_map_create(const uint64_t *pfn_array, uint32_t nr_pages)
{
    assert(pfn_array);

    hv_page_map *map = alloc_map(nr_pages);

    int i;
    for (i = 0; i < nr_pages; i++) {
        map->pg_slot[i].pfn = pfn_array[i];
    }

    return map;
}

void hv_page_map_destroy(hv_page_map *map)
{
    if (map) {
        int i;
        for (i = 0; i < map->nr_pages; ++i) {
            hv_page_map_unmap(map, i);
        }

        g_free(map);
    }
}

void hv_page_map_save(hv_page_map *map, QEMUFile *f)
{
    assert(map);
    assert(f);

    qemu_put_be32(f, map->nr_pages);

    int i;
    for (i = 0; i < map->nr_pages; ++i) {
        qemu_put_be64(f, map->pg_slot[i].pfn);
    }
}

hv_page_map *hv_page_map_load(QEMUFile *f)
{
    assert(f);

    uint32_t nr_pages = qemu_get_be32(f);
    hv_page_map *map = alloc_map(nr_pages);

    int i = 0;
    for (i = 0; i < nr_pages; ++i) {
        map->pg_slot[i].pfn = qemu_get_be64(f);
    }

    return map;
}
