/*
 *
 * Copyright (c) 2009, Microsoft Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307 USA.
 *
 * Authors:
 *   Haiyang Zhang <haiyangz@microsoft.com>
 *   Hank Janssen  <hjanssen@microsoft.com>
 *   K. Y. Srinivasan <kys@microsoft.com>
 *
 */

/*
 * VMBus ring buffer implementation
 * This ringbuffer format should be used by host and guest side
 * when communicating over vmbus channel.
 *
 * Ported from MS LIS code: driverv/hv/ring_buffer.c
 * Changes from main stream:
 * - Removed ring_lock - locking is unnessesary in our context
 * - On host side we have an array of guest PFNs which may not be contgious
 *   Thus raw buffer accesses are replaced with additional page map layer
 */

#include <qemu/osdep.h>
#include <qemu/atomic.h>
#include <target-i386/cpu.h>
#include "hw/vmbus/pagemap.h"
#include "hw/vmbus/ring_buffer.h"

/*
 *
 * hv_get_ringbuffer_availbytes()
 *
 * Get number of bytes available to read and to write to
 * for the specified ring buffer
 */
static inline void
hv_get_ringbuffer_availbytes(struct hv_ring_buffer_info *rbi,
              uint32_t *read, uint32_t *write)
{
    uint32_t read_loc, write_loc, dsize;

    smp_read_barrier_depends();

    /* Capture the read/write indices before they changed */
    read_loc = rbi->ring_buffer->read_index;
    write_loc = rbi->ring_buffer->write_index;
    dsize = rbi->ring_datasize;

    *write = write_loc >= read_loc ? dsize - (write_loc - read_loc) :
        read_loc - write_loc;
    *read = dsize - *write;
}

void hv_begin_read(struct hv_ring_buffer_info *rbi)
{
    rbi->ring_buffer->interrupt_mask = 1;
    smp_mb();
}

uint32_t hv_end_read(struct hv_ring_buffer_info *rbi)
{
    uint32_t read;
    uint32_t write;

    rbi->ring_buffer->interrupt_mask = 0;
    smp_mb();

    /*
     * Now check to see if the ring buffer is still empty.
     * If it is not, we raced and we need to process new
     * incoming messages.
     */
    hv_get_ringbuffer_availbytes(rbi, &read, &write);

    return read;
}

/*
 * When we write to the ring buffer, check if the host needs to
 * be signaled. Here is the details of this protocol:
 *
 *  1. The host guarantees that while it is draining the
 *     ring buffer, it will set the interrupt_mask to
 *     indicate it does not need to be interrupted when
 *     new data is placed.
 *
 *  2. The host guarantees that it will completely drain
 *     the ring buffer before exiting the read loop. Further,
 *     once the ring buffer is empty, it will clear the
 *     interrupt_mask and re-check to see if new data has
 *     arrived.
 */

static bool hv_need_to_signal(uint32_t old_write, struct hv_ring_buffer_info *rbi)
{
    smp_mb();
    if (rbi->ring_buffer->interrupt_mask)
        return false;

    /* check interrupt_mask before read_index */
    smp_rmb();
    /*
     * This is the only case we need to signal when the
     * ring transitions from being empty to non-empty.
     */
    if (old_write == rbi->ring_buffer->read_index)
        return true;

    return false;
}

/*
 * To optimize the flow management on the send-side,
 * when the sender is blocked because of lack of
 * sufficient space in the ring buffer, potential the
 * consumer of the ring buffer can signal the producer.
 * This is controlled by the following parameters:
 *
 * 1. pending_send_sz: This is the size in bytes that the
 *    producer is trying to send.
 * 2. The feature bit feat_pending_send_sz set to indicate if
 *    the consumer of the ring will signal when the ring
 *    state transitions from being full to a state where
 *    there is room for the producer to send the pending packet.
 */

static bool hv_need_to_signal_on_read(uint32_t old_rd,
                     struct hv_ring_buffer_info *rbi)
{
    uint32_t prev_write_sz;
    uint32_t cur_write_sz;
    uint32_t r_size;
    uint32_t write_loc = rbi->ring_buffer->write_index;
    uint32_t read_loc = rbi->ring_buffer->read_index;
    uint32_t pending_sz = rbi->ring_buffer->pending_send_sz;

    /*
     * If the other end is not blocked on write don't bother.
     */
    if (pending_sz == 0)
        return false;

    r_size = rbi->ring_datasize;
    cur_write_sz = write_loc >= read_loc ? r_size - (write_loc - read_loc) :
            read_loc - write_loc;

    prev_write_sz = write_loc >= old_rd ? r_size - (write_loc - old_rd) :
            old_rd - write_loc;


    if ((prev_write_sz < pending_sz) && (cur_write_sz >= pending_sz))
        return true;

    return false;
}

/*
 * hv_get_next_write_location()
 *
 * Get the next write location for the specified ring buffer
 *
 */
static inline uint32_t
hv_get_next_write_location(struct hv_ring_buffer_info *ring_info)
{
    uint32_t next = ring_info->ring_buffer->write_index;

    return next;
}

/*
 * hv_set_next_write_location()
 *
 * Set the next write location for the specified ring buffer
 *
 */
static inline void
hv_set_next_write_location(struct hv_ring_buffer_info *ring_info,
             uint32_t next_write_location)
{
    ring_info->ring_buffer->write_index = next_write_location;
}

/*
 * hv_get_next_read_location()
 *
 * Get the next read location for the specified ring buffer
 */
static inline uint32_t
hv_get_next_read_location(struct hv_ring_buffer_info *ring_info)
{
    uint32_t next = ring_info->ring_buffer->read_index;

    return next;
}

/*
 * hv_get_next_readlocation_withoffset()
 *
 * Get the next read location + offset for the specified ring buffer.
 * This allows the caller to skip
 */
static inline uint32_t
hv_get_next_readlocation_withoffset(struct hv_ring_buffer_info *ring_info,
                 uint32_t offset)
{
    uint32_t next = ring_info->ring_buffer->read_index;

    next += offset;
    next %= ring_info->ring_datasize;

    return next;
}

/*
 *
 * hv_set_next_read_location()
 *
 * Set the next read location for the specified ring buffer
 *
 */
static inline void
hv_set_next_read_location(struct hv_ring_buffer_info *ring_info,
            uint32_t next_read_location)
{
    ring_info->ring_buffer->read_index = next_read_location;
}

/*
 *
 * hv_get_ring_buffersize()
 *
 * Get the size of the ring buffer
 */
static inline uint32_t
hv_get_ring_buffersize(struct hv_ring_buffer_info *ring_info)
{
    return ring_info->ring_datasize;
}

/*
 *
 * hv_get_ring_bufferindices()
 *
 * Get the read and write indices as uint64_t of the specified ring buffer
 *
 */
static inline uint64_t
hv_get_ring_bufferindices(struct hv_ring_buffer_info *ring_info)
{
    return (uint64_t)ring_info->ring_buffer->write_index << 32;
}

/*
 *
 * hv_read_ringbuffer_data()
 *
 * Read data at proper offset inside page map
 *
 */
static inline void
hv_read_ringbuffer_data(struct hv_ring_buffer_info *ring_info, uint32_t pos,
        void* buf, uint32_t buflen)
{
    /* Page map ringbuffer data starts at page 1 */
    hv_page_map_read(ring_info->pmap, pos + TARGET_PAGE_SIZE, buf, buflen);
}

/*
 *
 * hv_write_ringbuffer_data()
 *
 * Write data at proper offset inside page map
 *
 */
static inline void
hv_write_ringbuffer_data(struct hv_ring_buffer_info *ring_info, uint32_t pos,
        void* buf, uint32_t buflen)
{
    /* Page map ringbuffer data starts at page 1 */
    hv_page_map_write(ring_info->pmap, pos + TARGET_PAGE_SIZE, buf, buflen);
}

/*
 *
 * hv_copyfrom_ringbuffer()
 *
 * Helper routine to copy to source from ring buffer.
 * Assume there is enough room. Handles wrap-around in src case only!!
 *
 */
static uint32_t hv_copyfrom_ringbuffer(
    struct hv_ring_buffer_info  *ring_info,
    void                *dest,
    uint32_t                destlen,
    uint32_t                start_read_offset)
{
    uint32_t ring_buffer_size = hv_get_ring_buffersize(ring_info);

    uint32_t frag_len;

    /* wrap-around detected at the src */
    if (destlen > ring_buffer_size - start_read_offset) {
        frag_len = ring_buffer_size - start_read_offset;

        hv_read_ringbuffer_data(ring_info, start_read_offset, dest, frag_len);
        hv_read_ringbuffer_data(ring_info, 0, dest + frag_len, destlen - frag_len);
    } else
        hv_read_ringbuffer_data(ring_info, start_read_offset, dest, destlen);


    start_read_offset += destlen;
    start_read_offset %= ring_buffer_size;

    return start_read_offset;
}


/*
 *
 * hv_copyto_ringbuffer()
 *
 * Helper routine to copy from source to ring buffer.
 * Assume there is enough room. Handles wrap-around in dest case only!!
 *
 */
static uint32_t hv_copyto_ringbuffer(
    struct hv_ring_buffer_info  *ring_info,
    uint32_t                start_write_offset,
    void                *src,
    uint32_t                srclen)
{
    uint32_t ring_buffer_size = hv_get_ring_buffersize(ring_info);
    uint32_t frag_len;

    /* wrap-around detected! */
    if (srclen > ring_buffer_size - start_write_offset) {
        frag_len = ring_buffer_size - start_write_offset;
        hv_write_ringbuffer_data(ring_info, start_write_offset, src, frag_len);
        hv_write_ringbuffer_data(ring_info, 0, src + frag_len, srclen - frag_len);
    } else
        hv_write_ringbuffer_data(ring_info, start_write_offset, src, srclen);

    start_write_offset += srclen;
    start_write_offset %= ring_buffer_size;

    return start_write_offset;
}

/*
 *
 * hv_ringbuffer_get_debuginfo()
 *
 * Get various debug metrics for the specified ring buffer
 *
 */
void hv_ringbuffer_get_debuginfo(struct hv_ring_buffer_info *ring_info,
                struct hv_ring_buffer_debug_info *debug_info)
{
    uint32_t bytes_avail_towrite;
    uint32_t bytes_avail_toread;

    if (ring_info->ring_buffer) {
        hv_get_ringbuffer_availbytes(ring_info,
                    &bytes_avail_toread,
                    &bytes_avail_towrite);

        debug_info->bytes_avail_toread = bytes_avail_toread;
        debug_info->bytes_avail_towrite = bytes_avail_towrite;
        debug_info->current_read_index =
            ring_info->ring_buffer->read_index;
        debug_info->current_write_index =
            ring_info->ring_buffer->write_index;
        debug_info->current_interrupt_mask =
            ring_info->ring_buffer->interrupt_mask;
    }
}

/*
 *
 * hv_ringbuffer_init()
 *
 *Initialize the ring buffer
 *
 */
int hv_ringbuffer_init(struct hv_ring_buffer_info *ring_info,
        const uint64_t *pfn_array, uint32_t nr_pages)
{
    if (!pfn_array || nr_pages < 1)
        return -EINVAL;

    if (sizeof(struct hv_ring_buffer) != TARGET_PAGE_SIZE)
        return -EINVAL;

    uint32_t buflen = nr_pages * TARGET_PAGE_SIZE;
    memset(ring_info, 0, sizeof(struct hv_ring_buffer_info));

    ring_info->pmap = hv_page_map_create(pfn_array, nr_pages);
    if (!ring_info->pmap)
        return -EFAULT;

    ring_info->ring_buffer = hv_page_map_map(ring_info->pmap, 0, true);
    if (!ring_info->ring_buffer) {
        hv_page_map_destroy(ring_info->pmap);
        return -EFAULT;
    }

    ring_info->ring_size = buflen;
    ring_info->ring_datasize = buflen - sizeof(struct hv_ring_buffer);

    return 0;
}

/*
 *
 * hv_ringbuffer_cleanup()
 *
 * Cleanup the ring buffer
 *
 */
void hv_ringbuffer_cleanup(struct hv_ring_buffer_info *ring_info)
{
    if (ring_info) {
        hv_page_map_destroy(ring_info->pmap);
        ring_info->ring_buffer = NULL;
        ring_info->pmap = NULL;
    }
}

/*
 *
 * hv_ringbuffer_write()
 *
 * Write to the ring buffer
 *
 */
int hv_ringbuffer_write(struct hv_ring_buffer_info *outring_info,
            struct hv_kvec *kv_list, uint32_t kv_count, bool *signal)
{
    int i = 0;
    uint32_t bytes_avail_towrite;
    uint32_t bytes_avail_toread;
    uint32_t totalbytes_towrite = 0;

    uint32_t next_write_location;
    uint32_t old_write;
    uint64_t prev_indices = 0;

    for (i = 0; i < kv_count; i++)
        totalbytes_towrite += kv_list[i].iov_len;

    totalbytes_towrite += sizeof(uint64_t);

    /*spin_lock_irqsave(&outring_info->ring_lock, flags);*/

    hv_get_ringbuffer_availbytes(outring_info,
                &bytes_avail_toread,
                &bytes_avail_towrite);


    /* If there is only room for the packet, assume it is full. */
    /* Otherwise, the next time around, we think the ring buffer */
    /* is empty since the read index == write index */
    if (bytes_avail_towrite <= totalbytes_towrite) {
        /*spin_unlock_irqrestore(&outring_info->ring_lock, flags);*/
        return -EAGAIN;
    }

    /* Write to the ring buffer */
    next_write_location = hv_get_next_write_location(outring_info);

    old_write = next_write_location;

    for (i = 0; i < kv_count; i++) {
        next_write_location = hv_copyto_ringbuffer(outring_info,
                             next_write_location,
                             kv_list[i].iov_base,
                             kv_list[i].iov_len);
    }

    /* Set previous packet start */
    prev_indices = hv_get_ring_bufferindices(outring_info);

    next_write_location = hv_copyto_ringbuffer(outring_info,
                         next_write_location,
                         &prev_indices,
                         sizeof(uint64_t));

    /* Issue a full memory barrier before updating the write index */
    smp_mb();

    /* Now, update the write location */
    hv_set_next_write_location(outring_info, next_write_location);


    /*spin_unlock_irqrestore(&outring_info->ring_lock, flags);*/

    *signal = hv_need_to_signal(old_write, outring_info);
    return 0;
}


/*
 *
 * hv_ringbuffer_peek()
 *
 * Read without advancing the read index
 *
 */
int hv_ringbuffer_peek(struct hv_ring_buffer_info *Inring_info,
           void *Buffer, uint32_t buflen)
{
    uint32_t bytes_avail_towrite;
    uint32_t bytes_avail_toread;
    uint32_t next_read_location = 0;

    /*spin_lock_irqsave(&Inring_info->ring_lock, flags);*/

    hv_get_ringbuffer_availbytes(Inring_info,
                &bytes_avail_toread,
                &bytes_avail_towrite);

    /* Make sure there is something to read */
    if (bytes_avail_toread < buflen) {

        /*spin_unlock_irqrestore(&Inring_info->ring_lock, flags);*/

        return -EAGAIN;
    }

    /* Convert to byte offset */
    next_read_location = hv_get_next_read_location(Inring_info);

    next_read_location = hv_copyfrom_ringbuffer(Inring_info,
                        Buffer,
                        buflen,
                        next_read_location);

    /*spin_unlock_irqrestore(&Inring_info->ring_lock, flags);*/

    return 0;
}

/*
 *
 * hv_ringbuffer_read()
 *
 * Read and advance the read index
 *
 */
int hv_ringbuffer_read(struct hv_ring_buffer_info *inring_info, void *buffer,
           uint32_t buflen, uint32_t offset, bool *signal)
{
    uint32_t bytes_avail_towrite;
    uint32_t bytes_avail_toread;
    uint32_t totalbytes_toread = 0;
    uint32_t next_read_location = 0;
    uint64_t prev_indices = 0;
    uint32_t old_read;

    if (buflen <= 0)
        return -EINVAL;

    /*spin_lock_irqsave(&inring_info->ring_lock, flags);*/

    hv_get_ringbuffer_availbytes(inring_info,
                &bytes_avail_toread,
                &bytes_avail_towrite);

    old_read = bytes_avail_toread;
    totalbytes_toread = buflen + sizeof(prev_indices);

    /* Make sure there is something to read */
    if (bytes_avail_toread < totalbytes_toread) {
        /*spin_unlock_irqrestore(&inring_info->ring_lock, flags);*/

        return -EAGAIN;
    }

    next_read_location =
        hv_get_next_readlocation_withoffset(inring_info, offset);

    next_read_location = hv_copyfrom_ringbuffer(inring_info,
                        buffer,
                        buflen,
                        next_read_location);

    next_read_location = hv_copyfrom_ringbuffer(inring_info,
                        &prev_indices,
                        sizeof(uint64_t),
                        next_read_location);

    /* Make sure all reads are done before we update the read index since */
    /* the writer may start writing to the read area once the read index */
    /*is updated */
    smp_mb();

    /* Update the read index */
    hv_set_next_read_location(inring_info, next_read_location);

    /*spin_unlock_irqrestore(&inring_info->ring_lock, flags);*/

    *signal = hv_need_to_signal_on_read(old_read, inring_info);

    return 0;
}

