/*
 * QEMU Hyper-V storage device support
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 *
 * TODO:
 * - Multichannel support
 */

#include "qemu/osdep.h"
#include "qemu/error-report.h"
#include "qapi/error.h"
#include "hw/vmbus/vmbus.h"
#include "standard-headers/asm-x86/hyperv.h"
#include "hw/scsi/scsi.h"
#include "sysemu/block-backend.h"
#include "sysemu/dma.h"
#include "block/scsi.h"

#define HV_SCSI_GUID "ba6163d9-04a1-4d29-b605-72e2ffb1dc7f"

#define VMSTOR_PROTO_VERSION(MAJOR_, MINOR_)	((((MAJOR_) & 0xff) << 8) | \
						(((MINOR_) & 0xff)))

#define VMSTOR_PROTO_VERSION_WIN6	VMSTOR_PROTO_VERSION(2, 0)
#define VMSTOR_PROTO_VERSION_WIN7	VMSTOR_PROTO_VERSION(4, 2)
#define VMSTOR_PROTO_VERSION_WIN8	VMSTOR_PROTO_VERSION(5, 1)
#define VMSTOR_PROTO_VERSION_WIN8_1	VMSTOR_PROTO_VERSION(6, 0)
#define VMSTAR_PROTO_VERSION_WIN10	VMSTOR_PROTO_VERSION(6, 2)

/*  Packet structure describing virtual storage requests. */
enum vstor_packet_operation {
	VSTOR_OPERATION_COMPLETE_IO		= 1,
	VSTOR_OPERATION_REMOVE_DEVICE		= 2,
	VSTOR_OPERATION_EXECUTE_SRB		= 3,
	VSTOR_OPERATION_RESET_LUN		= 4,
	VSTOR_OPERATION_RESET_ADAPTER		= 5,
	VSTOR_OPERATION_RESET_BUS		= 6,
	VSTOR_OPERATION_BEGIN_INITIALIZATION	= 7,
	VSTOR_OPERATION_END_INITIALIZATION	= 8,
	VSTOR_OPERATION_QUERY_PROTOCOL_VERSION	= 9,
	VSTOR_OPERATION_QUERY_PROPERTIES	= 10,
	VSTOR_OPERATION_ENUMERATE_BUS		= 11,
	VSTOR_OPERATION_FCHBA_DATA              = 12,
	VSTOR_OPERATION_CREATE_SUB_CHANNELS     = 13,
	VSTOR_OPERATION_MAXIMUM                 = 13
};

#define STORVSC_MAX_CMD_LEN			0x10
#define STORVSC_SENSE_BUFFER_SIZE		0x14
#define STORVSC_MAX_BUF_LEN_WITH_PADDING	0x14

enum storvsc_request_type {
    WRITE_TYPE = 0,
    READ_TYPE,
    UNKNOWN_TYPE,
};

struct vmscsi_win8_extension {
	/*
	 * The following were added in Windows 8
	 */
	uint16_t reserve;
	uint8_t  queue_tag;
	uint8_t  queue_action;
	uint32_t srb_flags;
	uint32_t time_out_value;
	uint32_t queue_sort_ey;
} __attribute((packed));

struct vmscsi_request {
	uint16_t length;
	uint8_t srb_status;
	uint8_t scsi_status;

	uint8_t  port_number;
	uint8_t  path_id;
	uint8_t  target_id;
	uint8_t  lun;

	uint8_t  cdb_length;
	uint8_t  sense_info_length;
	uint8_t  data_in;
	uint8_t  reserved;

	uint32_t data_transfer_length;

	union {
		uint8_t cdb[STORVSC_MAX_CMD_LEN];
		uint8_t sense_data[STORVSC_SENSE_BUFFER_SIZE];
		uint8_t reserved_array[STORVSC_MAX_BUF_LEN_WITH_PADDING];
	};
	/*
	 * The following was added in win8.
	 */
	struct vmscsi_win8_extension win8_extension;

} __attribute((packed));

/*
 * This structure is sent during the intialization phase to get the different
 * properties of the channel.
 */

#define STORAGE_CHANNEL_SUPPORTS_MULTI_CHANNEL		0x1

struct vmstorage_channel_properties {
	uint32_t reserved;
	uint16_t max_channel_cnt;
	uint16_t reserved1;

	uint32_t flags;
	uint32_t max_transfer_bytes;

	uint64_t reserved2;
} __attribute__((packed));

/*  This structure is sent during the storage protocol negotiations. */
struct vmstorage_protocol_version {
	/* Major (MSW) and minor (LSW) version numbers. */
	uint16_t major_minor;

	/*
	 * Revision number is auto-incremented whenever this file is changed
	 * (See FILL_VMSTOR_REVISION macro above).  Mismatch does not
	 * definitely indicate incompatibility--but it does indicate mismatched
	 * builds.
	 * This is only used on the windows side. Just set it to 0.
	 */
	uint16_t revision;
} __attribute__((packed));

/*
 * WWN packet for Fibre Channel HBA
 */

struct hv_fc_wwn_packet {
	uint8_t	primary_active;
	uint8_t	reserved1[3];
	uint8_t	primary_port_wwn[8];
	uint8_t	primary_node_wwn[8];
	uint8_t	secondary_port_wwn[8];
	uint8_t	secondary_node_wwn[8];
};

/*
 * SRB Flag Bits
 */

#define SRB_FLAGS_QUEUE_ACTION_ENABLE		0x00000002
#define SRB_FLAGS_DISABLE_DISCONNECT		0x00000004
#define SRB_FLAGS_DISABLE_SYNCH_TRANSFER	0x00000008
#define SRB_FLAGS_BYPASS_FROZEN_QUEUE		0x00000010
#define SRB_FLAGS_DISABLE_AUTOSENSE		0x00000020
#define SRB_FLAGS_DATA_IN			0x00000040
#define SRB_FLAGS_DATA_OUT			0x00000080
#define SRB_FLAGS_NO_DATA_TRANSFER		0x00000000
#define SRB_FLAGS_UNSPECIFIED_DIRECTION	(SRB_FLAGS_DATA_IN | SRB_FLAGS_DATA_OUT)
#define SRB_FLAGS_NO_QUEUE_FREEZE		0x00000100
#define SRB_FLAGS_ADAPTER_CACHE_ENABLE		0x00000200
#define SRB_FLAGS_FREE_SENSE_BUFFER		0x00000400

/*
 * This flag indicates the request is part of the workflow for processing a D3.
 */
#define SRB_FLAGS_D3_PROCESSING			0x00000800
#define SRB_FLAGS_IS_ACTIVE			0x00010000
#define SRB_FLAGS_ALLOCATED_FROM_ZONE		0x00020000
#define SRB_FLAGS_SGLIST_FROM_POOL		0x00040000
#define SRB_FLAGS_BYPASS_LOCKED_QUEUE		0x00080000
#define SRB_FLAGS_NO_KEEP_AWAKE			0x00100000
#define SRB_FLAGS_PORT_DRIVER_ALLOCSENSE	0x00200000
#define SRB_FLAGS_PORT_DRIVER_SENSEHASPORT	0x00400000
#define SRB_FLAGS_DONT_START_NEXT_PACKET	0x00800000
#define SRB_FLAGS_PORT_DRIVER_RESERVED		0x0F000000
#define SRB_FLAGS_CLASS_DRIVER_RESERVED		0xF0000000

struct vstor_packet {
	/* Requested operation type */
	enum vstor_packet_operation operation;

	/*  Flags - see below for values */
	uint32_t flags;

	/* Status of the request returned from the server side. */
	uint32_t status;

	/* Data payload area */
	union {
		/*
		 * Structure used to forward SCSI commands from the
		 * client to the server.
		 */
		struct vmscsi_request vm_srb;

		/* Structure used to query channel properties. */
		struct vmstorage_channel_properties storage_channel_properties;

		/* Used during version negotiations. */
		struct vmstorage_protocol_version version;

		/* Fibre channel address packet */
		struct hv_fc_wwn_packet wwn_packet;

		/* Number of sub-channels to create */
		uint16_t sub_channel_count;

		/* This will be the maximum of the union members */
		uint8_t  buffer[0x34];
	};
} __attribute__((packed));

/*
 * Packet Flags:
 *
 * This flag indicates that the server should send back a completion for this
 * packet.
 */

#define REQUEST_COMPLETION_FLAG	0x1

/*
 * SRB status codes and masks; a subset of the codes used here.
 */

#define SRB_STATUS_AUTOSENSE_VALID      0x80
#define SRB_STATUS_INVALID_LUN  0x20
#define SRB_STATUS_SUCCESS      0x01
#define SRB_STATUS_ABORTED      0x02
#define SRB_STATUS_ERROR        0x04

#define STORVSC_MAX_LUNS_PER_TARGET                     255
#define STORVSC_MAX_TARGETS                             2
#define STORVSC_MAX_CHANNELS                            8

#define STORVSC_FC_MAX_LUNS_PER_TARGET                  255
#define STORVSC_FC_MAX_TARGETS                          128
#define STORVSC_FC_MAX_CHANNELS                         8

#define STORVSC_SRB_MAX_TRANSFER_BYTES                  (2ul * 1024 * 1024)

////////////////////////////////////////////////////////////////////////////////

/**
 * VMBus SCSI controller
 */
typedef struct VMBusScsi {
    VMBusDevice parent;
    SCSIBus bus;
    enum {
        STORVSC_SCSI_RESET = 0,
        STORVSC_SCSI_INITIALIZING,
        STORVSC_SCSI_INITIALIZED,
    } state;
    uint8_t protocol_major;
    uint8_t protocol_minor;
} VMBusScsi;

#define STORVSC_SCSI_FROM_BUS(ptr) container_of(ptr, VMBusScsi, bus)

static const VMStateDescription storvsc_scsi_vmstate = {
    .name = TYPE_VMBUS_SCSI,
    .version_id = 1,
    .minimum_version_id = 1,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32(state, VMBusScsi),
        VMSTATE_UINT8(protocol_major, VMBusScsi),
        VMSTATE_UINT8(protocol_minor, VMBusScsi),
        VMSTATE_END_OF_LIST()
    }
};

/**
 * VMBus SCSI request
 */
typedef struct VMBusScsiReq
{
    VMBusScsi *s;

    SCSIRequest *sreq;
    struct vstor_packet msg;
    uint64_t request_id;

    hv_page_map *pmap;
    uint64_t data_offset;
    uint64_t data_bytes;
} VMBusScsiReq;

static void storvsc_scsi_reset(VMBusScsi *scsi)
{
    assert(scsi);

    qbus_reset_all(&scsi->bus.qbus);
    scsi->state = STORVSC_SCSI_RESET;
    scsi->protocol_major = 0;
    scsi->protocol_minor = 0;
}

static VMBusScsiReq *storvsc_scsi_create_req(VMBusScsi *s)
{
    VMBusScsiReq *req = g_malloc0(sizeof(*req));
    req->s = s;
    return req;
}

static void storvsc_scsi_free_req(VMBusScsiReq *req)
{
    hv_page_map_destroy(req->pmap);
    g_free(req);
}

/**
 * Packets we receive over channel
 * Full packet descriptor can be variable-sized in case of GPA_DIRECT packets
 */
struct storvsc_packet {
    struct vmpacket_descriptor desc;
    uint8_t data[];
};

/* Receive new packet over device channel.
 * Packet will be allocated on success, release it with storvsc_free_packet */
static struct storvsc_packet *storvsc_recv(VMBusDevice *dev)
{
    struct vmpacket_descriptor desc;
    int ret = vmbus_channel_peek(dev->chan, &desc);
    if (ret != 0) {
        VMBUS_TRACE("no packet: %d", ret);
        return NULL;
    }

    uint32_t len = vmpacket_get_length(&desc);
    struct storvsc_packet *msg = g_malloc(len);
    ret = vmbus_channel_recv_raw(dev->chan, msg, len, &len);
    if (ret != 0) {
        error_report("recv_raw failed: %d", ret);
        g_free(msg);
        return NULL;
    }

    assert(len == vmpacket_get_length(&desc));
    return msg;
}

/* Get pointer to packet data portion, skipping variable-sized descriptor */
static struct vstor_packet *storvsc_packet_get_data(struct storvsc_packet *msg)
{
    uint32_t offset = vmpacket_get_offset(&msg->desc);
    assert(offset >= sizeof(msg->desc));

    return (struct vstor_packet*)&msg->data[offset - sizeof(msg->desc)];
}

/* Get packet data portion size in bytes */
static uint32_t storvsc_packet_get_data_size(struct storvsc_packet *msg)
{
    return vmpacket_get_length(&msg->desc) - vmpacket_get_offset(&msg->desc);
}

static void storvsc_free_packet(struct storvsc_packet *msg)
{
    g_free(msg);
}

static void dump_vstor_packet(const struct vstor_packet *msg, uint32_t msglen)
{
    assert(msg);
    VMBUS_TRACE("operation: %d", msg->operation);
    VMBUS_TRACE("flags: %#x", msg->flags);
    VMBUS_TRACE("status: %d", msg->status);
}

static void dump_srb_packet(const struct vmscsi_request *srb)
{
    VMBUS_TRACE("length %d, port %d, path %d, target %d, lun %d",
            srb->length, srb->port_number, srb->path_id, srb->target_id, srb->lun);
    VMBUS_TRACE("srb_status %d, scsi_status %d",
            srb->srb_status, srb->scsi_status);
    VMBUS_TRACE("cdb size %d, sense info size %d",
            srb->cdb_length, srb->sense_info_length);
    VMBUS_TRACE("data_in = %d, data_transfer_length = %d",
            srb->data_in, srb->data_transfer_length);
    VMBUS_TRACE("CDB:");

    int i;
    for (i = 0; i < srb->cdb_length; ++i) {
        VMBUS_TRACE("[%d] 0x%hhx", i, srb->cdb[i]);
    }
}

static void storvsc_scsi_save_request(QEMUFile *f, SCSIRequest *sreq)
{
    VMBusScsiReq *req = (VMBusScsiReq*)sreq->hba_private;
    VMBUS_TRACE("saving request %p", req);

    /* vstor_packet is packed so we can save it as a buffer */
    qemu_put_buffer(f, (uint8_t*)&req->msg, sizeof(req->msg));
    qemu_put_be64(f, req->request_id);
    qemu_put_be64(f, req->data_offset);
    qemu_put_be64(f, req->data_bytes);
    hv_page_map_save(req->pmap, f);
}

static void *storvsc_scsi_load_request(QEMUFile *f, SCSIRequest *sreq)
{
    int res = 0;
    VMBusScsi *scsi = STORVSC_SCSI_FROM_BUS(sreq->bus);
    VMBusScsiReq *req = storvsc_scsi_create_req(scsi);

    res = qemu_get_buffer(f, (uint8_t*)&req->msg, sizeof(req->msg));
    if (res != 0) {
        storvsc_scsi_free_req(req);
        return NULL;
    }

    req->request_id = qemu_get_be64(f);
    req->data_offset = qemu_get_be64(f);
    req->data_bytes = qemu_get_be64(f);
    req->pmap = hv_page_map_load(f);
    if (!req->pmap) {
        storvsc_scsi_free_req(req);
        return NULL;
    }

    VMBUS_TRACE("loaded request %p", req);
    return req;
}

static void storvsc_scsi_request_cancelled(SCSIRequest *r)
{
    error_report("%s: not implemented", __FUNCTION__);
    assert(0);
}

static void storvsc_scsi_change(SCSIBus *bus, SCSIDevice *dev, SCSISense sense)
{
    error_report("%s: not implemented", __FUNCTION__);
    assert(0);
}

static int storvsc_scsi_parse_cdb(SCSIDevice *dev, SCSICommand *cmd,
                                 uint8_t *buf, void *hba_private)
{
    error_report("%s: not implemented", __FUNCTION__);
    assert(0);
    return 0;
}

static int complete_io(VMBusDevice *vdev, struct vstor_packet *msg,
        uint32_t msgsize, uint64_t request_id)
{
    msg->status = 0;
    msg->operation = VSTOR_OPERATION_COMPLETE_IO;
    return vmbus_channel_send(vdev->chan, msg, msgsize,
            VM_PKT_COMP, request_id);
}

static int storvsc_scsi_complete_req(VMBusScsiReq *req)
{
    assert(req);
    return complete_io(&req->s->parent, &req->msg, sizeof(req->msg), req->request_id);
}

static void do_transfer(VMBusScsiReq *req, uint8_t *buf, uint32_t len, int dir)
{
    assert(req->pmap);
    assert(len <= req->data_bytes);

    bool ret = false;
    if (dir == SCSI_XFER_FROM_DEV) {
        ret = hv_page_map_write(req->pmap, req->data_offset, buf, len);
    } else if (dir == SCSI_XFER_TO_DEV) {
        ret = hv_page_map_read(req->pmap, req->data_offset, buf, len);
    } else {
        assert(0);
    }

    if (!ret) {
        error_report("data transfer failed");
        return;
    }

    req->data_offset += len;
}

static void storvsc_scsi_transfer_data(SCSIRequest *r, uint32_t len)
{
    VMBUS_TRACE("(%p, %d)", r, len);
    VMBusScsiReq *req = (VMBusScsiReq*)r->hba_private;

    uint8_t* buf = scsi_req_get_buf(r);
    VMBUS_TRACE("buf %p", (void*)buf);

    do_transfer(req, buf, len, r->cmd.mode);
    scsi_req_continue(r);
}

static void storvsc_scsi_command_complete(SCSIRequest *r, uint32_t status,
                                         size_t resid)
{
    VMBUS_TRACE("(%p, %d, %zu)", r, status, resid);
    VMBusScsiReq *req = (VMBusScsiReq*)r->hba_private;

    req->msg.vm_srb.scsi_status = status;
    req->msg.vm_srb.srb_status = SRB_STATUS_SUCCESS;

    uint32_t sense_len = scsi_req_get_sense(r, req->msg.vm_srb.sense_data, sizeof(req->msg.vm_srb.sense_data));
    assert(sense_len <= sizeof(req->msg.vm_srb.sense_data));
    VMBUS_TRACE("sense_len = %d (%d)", sense_len, r->sense_len);
    req->msg.vm_srb.sense_info_length = sense_len;

    if (sense_len > 0) {
        req->msg.vm_srb.srb_status |= SRB_STATUS_AUTOSENSE_VALID;
        VMBUS_TRACE("sense:");

        int i = 0;
        for (;i < sense_len; ++i) {
            VMBUS_TRACE("[%d] 0x%hhx", i, req->msg.vm_srb.sense_data[i]);
        }
    }

    assert(resid <= req->msg.vm_srb.data_transfer_length);
    req->msg.vm_srb.data_transfer_length -= resid;

    storvsc_scsi_complete_req(req);
    storvsc_scsi_free_req(req);
}


static struct SCSIBusInfo storvsc_scsi_info = {
    .tcq = true,
    .max_channel = STORVSC_MAX_CHANNELS,
    .max_target = STORVSC_MAX_TARGETS,
    .max_lun = STORVSC_MAX_LUNS_PER_TARGET - 1,
    .complete = storvsc_scsi_command_complete,
    .cancel = storvsc_scsi_request_cancelled,
    .change = storvsc_scsi_change,
    .parse_cdb = storvsc_scsi_parse_cdb,
    .transfer_data = storvsc_scsi_transfer_data,
    .save_request = storvsc_scsi_save_request,
    .load_request = storvsc_scsi_load_request,
};

/* TODO: figure out a better way to handle bad targets */
static void handle_missing_target(VMBusScsiReq *req)
{
    struct vmscsi_request *srb = &req->msg.vm_srb;
    srb->srb_status = SRB_STATUS_INVALID_LUN | SRB_STATUS_SUCCESS;
    srb->scsi_status = 0;

    switch (srb->cdb[0]) {
        uint8_t data;
        uint32_t data32;
        case INQUIRY:
            VMBUS_TRACE("INQUIRY reply for bad target");
            data = 0x7F;
            do_transfer(req, &data, 1, SCSI_XFER_FROM_DEV);
            break;

        case REPORT_LUNS:
            VMBUS_TRACE("REPORT LUNS reply for bad target");
            data32 = 0;
            do_transfer(req, (uint8_t*)&data32, 4, SCSI_XFER_FROM_DEV);
        default:
            srb->scsi_status = CHECK_CONDITION;
            srb->srb_status |= SRB_STATUS_AUTOSENSE_VALID;
            srb->data_transfer_length -= 4;
            memset(srb->sense_data, 0, sizeof(srb->sense_info_length));
            srb->sense_data[0] = 0x70;
            srb->sense_data[2] = 0x05;
            srb->sense_data[7] = srb->sense_info_length - 7;
            srb->sense_data[12] = 0x25;
            srb->sense_data[13] = 0x00;
            break;
    }

    storvsc_scsi_complete_req(req);
}

static void storvsc_scsi_execute_srb(VMBusScsi *scsi, struct storvsc_packet *packet)
{
    assert(scsi);
    assert(packet);

    struct vstor_packet *msg = storvsc_packet_get_data(packet);
    struct vmscsi_request *srb = &msg->vm_srb;
    dump_srb_packet(srb);

    VMBusScsiReq *req = storvsc_scsi_create_req(scsi);
    req->msg = *msg;
    req->request_id = packet->desc.trans_id;

    if (packet->desc.type == VM_PKT_DATA_USING_GPA_DIRECT) {
        struct vmpacket_mpb *mpb = (struct vmpacket_mpb*)packet;
        uint32_t pages = DIV_ROUND_UP(mpb->range.byte_offset + mpb->range.byte_count, TARGET_PAGE_SIZE);
        VMBUS_TRACE("GPA pages %u, bytes %d, offset %d", pages, mpb->range.byte_count, mpb->range.byte_offset);

        int i;
        for (i = 0; i < pages; ++i) {
            VMBUS_TRACE("GPA %d = 0x%llx", i, (unsigned long long)mpb->range.pfn[i]);
        }

        req->pmap = hv_page_map_create(mpb->range.pfn, pages);
        req->data_bytes = mpb->range.byte_count;
        req->data_offset = mpb->range.byte_offset;
    }

    SCSIDevice *d = scsi_device_find(&scsi->bus, srb->path_id, srb->target_id, srb->lun);
    if (!d) {
        VMBUS_TRACE("scsi device not found");
        handle_missing_target(req);
        return;
    }

    VMBUS_TRACE("SCSIDevice %p: channel %d, target %d, lun %d", d, d->channel, d->id, d->lun);

    req->sreq = scsi_req_new(d, srb->path_id, srb->lun, srb->cdb, req);
    assert(req->sreq);

    VMBUS_TRACE("Created scsi request %p", req->sreq);

    scsi_req_ref(req->sreq);
    blk_io_plug(d->conf.blk);
    if (scsi_req_enqueue(req->sreq)) {
        scsi_req_continue(req->sreq);
    }
    blk_io_unplug(req->sreq->dev->conf.blk);
    scsi_req_unref(req->sreq);
}

static void storvsc_scsi_signal_event(EventNotifier *e)
{
    if (!event_notifier_test_and_clear(e)) {
        return;
    }

    VMBusDevice *dev = container_of(e, VMBusDevice, signal_event);
    VMBusScsi *scsi = VMBUS_SCSI(dev);

    VMBUS_TRACE("%p", dev);

    do {
        bool do_complete = true;

        struct storvsc_packet *packet = storvsc_recv(dev);
        if (!packet) {
            break;
        }

        uint32_t msglen = storvsc_packet_get_data_size(packet);
        struct vstor_packet *msg = storvsc_packet_get_data(packet);
        dump_vstor_packet(msg, msglen);

        struct vstor_packet reply;
        memset(&reply, 0, sizeof(reply));

        if (scsi->state == STORVSC_SCSI_RESET) {
            switch (msg->operation) {
                case VSTOR_OPERATION_BEGIN_INITIALIZATION:
                    scsi->state = STORVSC_SCSI_INITIALIZING;
                    break;

                default:
                    error_report("Invalid operation %d for state %d", msg->operation, scsi->state);
                    assert(0);
            }
        } else if (scsi->state == STORVSC_SCSI_INITIALIZING) {
            switch (msg->operation) {
                case VSTOR_OPERATION_QUERY_PROTOCOL_VERSION:
                    scsi->protocol_major = msg->version.major_minor >> 8;
                    scsi->protocol_minor = msg->version.major_minor & 0xFF;
                    VMBUS_TRACE("client version %d.%d",
                            scsi->protocol_major, scsi->protocol_minor);
                    break;

                case VSTOR_OPERATION_QUERY_PROPERTIES:
                    reply.storage_channel_properties.max_channel_cnt = 1;
                    reply.storage_channel_properties.flags = 0;
                    reply.storage_channel_properties.max_transfer_bytes = STORVSC_SRB_MAX_TRANSFER_BYTES;
                    break;

                case VSTOR_OPERATION_END_INITIALIZATION:
                    scsi->state = STORVSC_SCSI_INITIALIZED;
                    break;

                default:
                    error_report("Invalid operation %d for state %d", msg->operation, scsi->state);
                    assert(0);
            }
        } else if (scsi->state == STORVSC_SCSI_INITIALIZED) {
            switch (msg->operation) {
                case VSTOR_OPERATION_EXECUTE_SRB:
                    storvsc_scsi_execute_srb(scsi, packet);
                    do_complete = false;
                    break;

                default:
                    error_report("unknown vstor packet operation %d", msg->operation);
                    assert(0);
            }
        } else {
            assert(0);
        }

        if (do_complete && (msg->flags & REQUEST_COMPLETION_FLAG)) {
            /* Windows clients want us to reply with a packet with the same length
             * as was its request packet, which can be 60 or 48 bytes but has basically
             * same structure. Otherwise windows does not recognize our scsi drives. */
            complete_io(dev, &reply, msglen, packet->desc.trans_id);
        }

        storvsc_free_packet(packet);
    } while (1);

    VMBUS_TRACE("Completed packet handling");
}

///////////////////////////////////////////////////////////////////////////////

static int storvsc_scsi_dev_open_channel(VMBusDevice *dev, unsigned char *userdata)
{
    VMBUS_TRACE("%p", dev);
    return 0;
}

static void storvsc_scsi_dev_close_channel(VMBusDevice *dev)
{
    VMBusScsi *scsi = VMBUS_SCSI(dev);
    VMBUS_TRACE("%p", dev);

    storvsc_scsi_reset(scsi);
}

static void storvsc_scsi_dev_realize(DeviceState *dev, Error **errp)
{
    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusScsi *scsi = VMBUS_SCSI(dev);
    Error *local_err = NULL;
    uint16_t chn_flags = VMBUS_CHANNEL_SERVER_SUPPORTS_TRANSFER_PAGES;

    VMBUS_TRACE("%p", dev);

    vdev->chan = vmbus_create_channel(vdev, chn_flags);

    if (!vdev->chan) {
        error_setg(&local_err, "failed to create channel");
        goto error_out;
    }

    int res = event_notifier_init(&vdev->signal_event, 0);
    if (res != 0) {
        error_setg(&local_err, "event notifier failed to init with %d", res);
        goto error_out;
    }

    res = event_notifier_set_handler(&vdev->signal_event, false, storvsc_scsi_signal_event);
    if (res != 0) {
        error_setg(&local_err, "event notifier set handler failed with %d", res);
        goto error_out;
    }

    res = vmstate_register(dev, -1, &storvsc_scsi_vmstate, dev);
    if (res != 0) {
        error_setg(&local_err, "vmstate_register says %d", res);
        goto error_out;
    }

    scsi_bus_new(&scsi->bus, sizeof(scsi->bus), (DeviceState*)scsi,
            &storvsc_scsi_info, NULL);
    return;

error_out:
    vmbus_release_channel(vdev->chan);
    event_notifier_cleanup(&vdev->signal_event);
    error_propagate(errp, local_err);
}

static void storvsc_scsi_dev_reset(DeviceState *dev)
{
    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusScsi *scsi = VMBUS_SCSI(vdev);

    VMBUS_TRACE("%p", scsi);

    storvsc_scsi_reset(scsi);
}

static void storvsc_scsi_dev_unrealize(DeviceState *dev, Error **errp)
{
    VMBUS_TRACE("%p", dev);
}

static void storvsc_scsi_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);
    VMBusDeviceClass *vdc = VMBUS_DEVICE_CLASS(klass);

    int res = qemu_uuid_parse(HV_SCSI_GUID, &vdc->class_uuid);
    assert(res == 0);

    dc->fw_name = "scsi";
    vdc->realize = storvsc_scsi_dev_realize;
    vdc->unrealize = storvsc_scsi_dev_unrealize;
    vdc->reset = storvsc_scsi_dev_reset;
    vdc->open_channel = storvsc_scsi_dev_open_channel;
    vdc->close_channel = storvsc_scsi_dev_close_channel;
}

static const TypeInfo storvsc_scsi_type_info = {
    .name = TYPE_VMBUS_SCSI,
    .parent = TYPE_VMBUS_DEVICE,
    .instance_size = sizeof(VMBusScsi),
    .class_init = storvsc_scsi_class_init,
};

static void storvsc_register_types(void)
{
    type_register_static(&storvsc_scsi_type_info);
}

type_init(storvsc_register_types)
