/*
 * QEMU Hyper-V VMBus support
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 */

#include "qemu/osdep.h"
#include "qemu/error-report.h"
#include "qemu/coroutine.h"
#include "qapi/error.h"
#include "hw/vmbus/vmbus.h"
#include "hw/sysbus.h"
#include "target-i386/hyperv_defs.h"

#define VMBUS_STATE(obj) OBJECT_CHECK(VMBus, (obj), TYPE_VMBUS)
#define VMBUS_DEVICE(obj) OBJECT_CHECK(VMBusDevice, (obj), TYPE_VMBUS_DEVICE)

#define VMBUS_SINT_ROUTE_NAME "vmbus-sint-route"
#define VMBUS_RX_QUEUE_CAPACITY 16

/*
 * Every channel id is used as an event flag bit number in event page.
 * Thus maximum number of channel resource ids is limited
 * by a number of available signal event bits.
 */
#define VMBUS_CHAN_MAX_RELID    HV_EVENT_FLAGS_COUNT
#define VMBUS_CHAN_FIRST_RELID  1 /* 0 is reserved */
#define VMBUS_CHAN_MAX_CONNID   HV_EVENT_FLAGS_COUNT
#define VMBUS_CHAN_FIRST_CONNID (VMBUS_MONITOR_CONNECTION_ID + 1)

#define VMBUS_VMSTATE_VERSION_ID 1

/* Save single struct field as a buffer */
#define VMSTATE_SINGLE_BUFFER(_field, _state) \
    VMSTATE_SINGLE(_field, _state, 1, vmstate_info_buffer, typeof_field(_state, _field))

/**
 * Valid vmbus states
 * State numberical order matters
 */
typedef enum VMBusState {
    /** Initial state */
    VMBusState_Reset = 0,

    /** VMBus can send replies and signal events to guest */
    VMBusState_Initialized,

    /** CHANNELMSG_INITIATE_CONTACT handled */
    VMBusState_ContactInitiated,

    /** All channel offers have been delivered */
    VMBusState_OffersDelivered,
} VMBusState;

typedef void(*VMBusReplyCompletionHandler)(struct vmbus_channel_message_header *msg, int res);

/**
 * VMBus context
 */
typedef struct VMBus
{
    BusState parent;

    /* Current state */
    VMBusState state;

    /* Copy of initial guest handshake message */
    struct vmbus_channel_initiate_contact initiate_contact;

    /* CPU that guest wants to receive our reply messages on */
    X86CPU* target_vcpu;

    /* Prepared vmbus sint route for target vcpu */
    HvSintRoute *sint_route;

    /* Mapped interrupt page if guest wants to use it for channel events */
    struct hv_interrupt_page *int_page;

    /* Allocated chanid bitmap */
    unsigned long chan_relid_bitmap[BITS_TO_LONGS(VMBUS_CHAN_MAX_RELID)];

    /* Channel connection id bitmap */
    unsigned long connection_id_bitmap[BITS_TO_LONGS(VMBUS_CHAN_MAX_CONNID)];

    /* Incoming message queue.
     * Posted messages are dumped from VCPU threads here.
     * Lock protects concurrent access from multiple VCPU threads.
     * BH is scheduled to signal new messages for processing in main loop. */
    struct hv_input_post_message rx_queue[VMBUS_RX_QUEUE_CAPACITY];
    size_t rx_queue_head;
    size_t rx_queue_size;
    QemuMutex rx_queue_lock;
    QEMUBH *bh;
    bool has_posted_message;

    QLIST_HEAD(, VMBusGpadl) gpadl_list;
    QLIST_HEAD(, VMBusChannel) channel_list;
} VMBus;

static VMBus *g_vmbus;

static void sint_callback(HvSintRoute *sint_route);
static int post_reply(struct vmbus_channel_message_header *msg, uint32_t size);

////////////////////////////////////////////////////////////////////////////////

static void *map_guest_page(hwaddr addr, bool is_write)
{
    hwaddr len = 1;
    void *pageaddr = cpu_physical_memory_map(addr, &len, is_write);
    if (!pageaddr || (len != 1)) {
        return NULL;
    }

    return pageaddr;
}

static void unmap_guest_page(void* addr, bool is_write)
{
    if (addr) {
        hwaddr len = 1;
        cpu_physical_memory_unmap(addr, len, is_write, len);
    }
}

static HvSintRoute *make_vmbus_sint_route(X86CPU *cpu)
{
    HvSintRoute *route = NULL;
    route = hyperv_sint_route_get(cpu, VMBUS_SINT_ROUTE_NAME);
    if (!route) {
        VMBUS_TRACE("No existing vmbus sint route for vcpu %p", cpu);
        route = hyperv_sint_route_new(cpu, VMBUS_MESSAGE_SINT,
                VMBUS_SINT_ROUTE_NAME, sint_callback);
    }

    return route;
}

static const VMStateDescription vmbus_vmstate_gpadl = {
    .name = "vmbus/gpadl",
    .version_id = 1,
    .minimum_version_id = 1,
    .fields = (VMStateField[]) {
        VMSTATE_BOOL(created, struct VMBusGpadl),
        VMSTATE_UINT32(id, struct VMBusGpadl),
        VMSTATE_UINT32(child_relid, struct VMBusGpadl),
        VMSTATE_UINT32(total_pages, struct VMBusGpadl),
        VMSTATE_UINT32(seen_pages, struct VMBusGpadl),
        {
            .name       = "gfn",
            .version_id = 1,
            .num_offset = vmstate_offset_value(struct VMBusGpadl, total_pages, uint32_t),
            .info       = &vmstate_info_buffer,
            .size       = sizeof(uint64_t),
            .flags      = VMS_VARRAY_UINT32,
            .offset     = offsetof(struct VMBusGpadl, gfn),
        },
        VMSTATE_END_OF_LIST()
    }
};

static VMBusGpadl *alloc_gpadl(uint32_t total_pages)
{
    size_t gpadl_size = sizeof(VMBusGpadl) + (total_pages * sizeof(uint64_t));
    return g_malloc(gpadl_size);
}

static VMBusGpadl *create_gpadl(uint32_t gpadl_id, uint32_t child_relid,
        uint32_t size)
{
    assert((size & (TARGET_PAGE_SIZE - 1)) == 0);

    uint32_t gfns = size >> TARGET_PAGE_BITS;
    VMBusGpadl *gpadl = alloc_gpadl(gfns);

    gpadl->id = gpadl_id;
    gpadl->child_relid = child_relid;
    gpadl->total_pages = gfns;
    gpadl->seen_pages = 0;
    gpadl->created = false;

    QLIST_INSERT_HEAD(&g_vmbus->gpadl_list, gpadl, link);
    return gpadl;
}

static void save_gpadl(QEMUFile *f, VMBusGpadl *gpadl)
{
    assert(f);
    assert(gpadl);

    qemu_put_be32(f, gpadl->total_pages);
    vmstate_save_state(f, &vmbus_vmstate_gpadl, gpadl, NULL);
}

static int load_gpadl(QEMUFile *f)
{
    assert(f);

    uint32_t pages = qemu_get_be32(f);
    assert(pages != 0);

    VMBusGpadl *gpadl = alloc_gpadl(pages);
    int ret = vmstate_load_state(f, &vmbus_vmstate_gpadl, gpadl,
            VMBUS_VMSTATE_VERSION_ID);
    if (ret != 0) {
        g_free(gpadl);
        return ret;
    }

    QLIST_INSERT_HEAD(&g_vmbus->gpadl_list, gpadl, link);
    return 0;
}

static VMBusGpadl *find_gpadl(uint32_t gpadl_id)
{
    VMBusGpadl *gpadl = NULL;
    QLIST_FOREACH(gpadl, &g_vmbus->gpadl_list, link) {
        if (gpadl->id == gpadl_id) {
            return gpadl;
        }
    }

    return NULL;
}

static void free_gpadl(VMBusGpadl *gpadl)
{
    if (gpadl) {
        QLIST_REMOVE(gpadl, link);
        g_free(gpadl);
    }
}

static uint32_t alloc_chan_relid(void)
{
    /* relid 0 is special case in guest vmbus so skip this value */
    uint32_t i;
    for (i = VMBUS_CHAN_FIRST_RELID; i < VMBUS_CHAN_MAX_RELID; ++i) {
        if (!test_and_set_bit(i, g_vmbus->chan_relid_bitmap)) {
            return i;
        }
    }

    return 0;
}

static void free_chan_relid(uint32_t relid)
{
    if (relid > VMBUS_CHAN_MAX_RELID) {
        return;
    }

    clear_bit(relid, g_vmbus->chan_relid_bitmap);
}

static uint32_t alloc_connection_id(void)
{
    uint32_t i;
    for (i = VMBUS_CHAN_FIRST_CONNID; i < VMBUS_CHAN_MAX_CONNID; ++i) {
        if (!test_and_set_bit(i, g_vmbus->connection_id_bitmap)) {
            return i;
        }
    }

    return 0;
}

static void free_connection_id(uint32_t connid)
{
    if (connid > VMBUS_CHAN_MAX_CONNID) {
        return;
    }

    clear_bit(connid, g_vmbus->connection_id_bitmap);
}

static VMBusChannel *alloc_channel(void)
{
    VMBusChannel *chan = g_malloc0(sizeof(*chan));
    QLIST_INSERT_HEAD(&g_vmbus->channel_list, chan, link);
    return chan;
}

/* Release channel resources but don't free it
 * This function handles partially opened channels to cleanup failed opens */
static void close_channel(VMBusChannel *chan)
{
    if (chan) {
        VMBusDeviceClass *vdc = VMBUS_DEVICE_GET_CLASS(chan->owner);
        if (vdc->close_channel != NULL) {
            vdc->close_channel(chan->owner);
        }

        hyperv_sint_route_unref(chan->notify_route);
        chan->notify_route = NULL;

        hv_ringbuffer_cleanup(&chan->rb_inbound);
        hv_ringbuffer_cleanup(&chan->rb_outbound);
        chan->is_open = false;
        chan->offer_sent = false;
    }
}

static void free_channel(VMBusChannel *chan)
{
    if (chan) {
        if (chan->is_open) {
            close_channel(chan);
        }

        QLIST_REMOVE(chan, link);
        g_free(chan);
    }
}

static int init_channel_common(VMBusChannel *chan)
{
    int res = 0;
    assert(chan);

    X86CPU *vcpu = hyperv_find_vcpu(chan->target_vcpu);
    if (!vcpu) {
        return -EINVAL;
    }

    VMBusGpadl *gpadl = find_gpadl(chan->rb_gpadl);
    if (!gpadl || !gpadl->created) {
        return -EINVAL;
    }

    if (chan->rb_page_offset >= gpadl->total_pages) {
        return -EINVAL;
    }

    if (!(chan->rb_page_offset > 0 && chan->rb_page_offset < gpadl->total_pages)) {
        return -EINVAL;
    }

    chan->notify_route = make_vmbus_sint_route(vcpu);
    if (!chan->notify_route) {
        res = -EINVAL;
        goto error_out;
    }

    /*
     * Channel ringbuffer is divided by rb_offset into two areas in PFN array
     * [0 - rb_offset) - outbound ringbuffer, host-writable
     * [rb_offset - total_pages) - inbound ringbuffer, host-readable
     */

    res = hv_ringbuffer_init(&chan->rb_inbound, gpadl->gfn, chan->rb_page_offset);
    if (res) {
        goto error_out;
    }

    res = hv_ringbuffer_init(&chan->rb_outbound, &gpadl->gfn[chan->rb_page_offset],
                gpadl->total_pages - chan->rb_page_offset);
    if (res) {
        goto error_out;
    }

    return 0;

error_out:
    hyperv_sint_route_unref(chan->notify_route);
    hv_ringbuffer_cleanup(&chan->rb_inbound);
    hv_ringbuffer_cleanup(&chan->rb_outbound);
    return res;
}

static int open_channel(VMBusChannel *chan,
        struct vmbus_channel_open_channel *msg)
{
    int res = 0;

    assert(chan);
    assert(msg);

    if (chan->is_open) {
        return -EINVAL;
    }

    chan->rb_gpadl = msg->ringbuffer_gpadlhandle;
    chan->rb_page_offset = msg->downstream_ringbuffer_pageoffset;
    chan->target_vcpu = msg->target_vp;

    res = init_channel_common(chan);
    if (res != 0) {
        return res;
    }

    VMBusDeviceClass *vdc = VMBUS_DEVICE_GET_CLASS(chan->owner);
    if (vdc->open_channel != NULL) {
        res = vdc->open_channel(chan->owner, msg->userdata);
        if (res != 0) {
            close_channel(chan);
            return res;
        }
    }

    chan->is_open = true;
    return 0;
}

/* Restore channel state after migration */
static int load_channel(VMBusChannel *chan)
{
    assert(chan);

    if (!chan->is_open) {
        return 0;
    }

    return init_channel_common(chan);
}

static int vmbus_channel_post_load(void *opaque, int version_id)
{
    assert(opaque);
    VMBusChannel *chan = (VMBusChannel*)opaque;
    return load_channel(chan);
}

/* struct VMBusChannel */
static const VMStateDescription vmbus_vmstate_channel = {
    .name = "vmbus/channel",
    .version_id = 1,
    .minimum_version_id = 1,
    .post_load = vmbus_channel_post_load,
    .fields = (VMStateField[]) {
        VMSTATE_SINGLE_BUFFER(offer, struct VMBusChannel),
        VMSTATE_BOOL(offer_sent, struct VMBusChannel),
        VMSTATE_BOOL(is_open, struct VMBusChannel),
        VMSTATE_UINT32(relid, struct VMBusChannel),
        VMSTATE_UINT32(connection_id, struct VMBusChannel),
        VMSTATE_UINT32(target_vcpu, struct VMBusChannel),
        VMSTATE_UINT32(rb_gpadl, struct VMBusChannel),
        VMSTATE_UINT32(rb_page_offset, struct VMBusChannel),
        VMSTATE_END_OF_LIST()
    }
};

static VMBusChannel *find_channel(uint32_t relid)
{
    VMBusChannel *chan = NULL;
    QLIST_FOREACH(chan, &g_vmbus->channel_list, link) {
        if (chan->relid == relid) {
            return chan;
        }
    }

    return NULL;
}

static VMBusChannel *find_channel_connid(uint32_t connid)
{
    VMBusChannel *chan = NULL;
    QLIST_FOREACH(chan, &g_vmbus->channel_list, link) {
        if (chan->connection_id == connid) {
            return chan;
        }
    }

    return NULL;
}

static int signal_channel(VMBusChannel *chan)
{
    int res;

    assert(chan);

    X86CPU *vcpu = hyperv_find_vcpu(chan->target_vcpu);
    if (!vcpu) {
        error_report("no vcpu id %d", chan->target_vcpu);
        return -EINVAL;
    }

    /*
     * Pre-win8 guests want to see channel relid in interrupt page
     * Other guests want same channel relid in hyperv event page
     */

    if (g_vmbus->int_page) {
        struct hv_interrupt_page *int_page = g_vmbus->int_page;
        assert(int_page != NULL);

        int i = chan->relid / BITS_PER_LONG;
        unsigned long mask = 1ul << (chan->relid % BITS_PER_LONG);
        if ((atomic_fetch_or(&int_page->to_guest[i], mask) & mask) != mask) {
            res = kvm_hv_sint_route_set_sint(chan->notify_route);
        } else {
            res = 0;
        }
    } else {
        res = hyperv_set_evt_flag(chan->notify_route, chan->relid);
        if (res != 0 && res != -ENXIO) {
            error_report("failed to set channel event flag");
        }
    }

    return res;
}

/* Get rx queue array index based on logical position */
static inline size_t get_queue_index(size_t pos)
{
    return (g_vmbus->rx_queue_head + pos) % VMBUS_RX_QUEUE_CAPACITY;
}

static bool have_incoming_message(void)
{
    return g_vmbus->rx_queue_size != 0;
}

static bool dequeue_incoming_message(struct hv_input_post_message *msg)
{
    /* RX dequeue is done in IO thread which is locked by global io mutex
     * So no data races with VCPU threads and no locking needed */
    if (g_vmbus->rx_queue_size == 0) {
        return false;
    }

    memcpy(msg, &g_vmbus->rx_queue[g_vmbus->rx_queue_head], sizeof(*msg));
    g_vmbus->rx_queue_head = (g_vmbus->rx_queue_head + 1)
        % VMBUS_RX_QUEUE_CAPACITY;
    g_vmbus->rx_queue_size--;
    return true;
}

static bool enqueue_incoming_message(const struct hv_input_post_message *msg)
{
    /* Locks protect against concurrent VCPU thread enqueue
     * (see comments in dequeue) */
    qemu_mutex_lock(&g_vmbus->rx_queue_lock);

    if (g_vmbus->rx_queue_size == VMBUS_RX_QUEUE_CAPACITY) {
        qemu_mutex_unlock(&g_vmbus->rx_queue_lock);
        return false;
    }

    size_t index = (g_vmbus->rx_queue_head + g_vmbus->rx_queue_size)
        % VMBUS_RX_QUEUE_CAPACITY;
    memcpy(&g_vmbus->rx_queue[index], msg, sizeof(*msg));
    g_vmbus->rx_queue_size++;

    qemu_mutex_unlock(&g_vmbus->rx_queue_lock);
    qemu_bh_schedule(g_vmbus->bh);
    return true;
}

/* Transitions state from Reset to Initialized */
static int vmbus_init_with_initiate_contact(
        struct vmbus_channel_initiate_contact *msg)
{
    assert(msg);
    assert(g_vmbus->state == VMBusState_Reset);

    X86CPU *target_vcpu = hyperv_find_vcpu(msg->target_vcpu);
    if (!target_vcpu) {
        return -EINVAL;
    }

    void *int_page = NULL;
    if (msg->vmbus_version_requested < VMBUS_VERSION_WIN8) {
        if (msg->interrupt_page == 0) {
            return -EINVAL;
        }

        int_page = map_guest_page(msg->interrupt_page, true);
        if (!int_page) {
            return -ENOMEM;
        }
    }

    HvSintRoute *route = make_vmbus_sint_route(target_vcpu);
    if (!route) {
        unmap_guest_page(int_page, true);
        return -ENXIO;
    }

    g_vmbus->initiate_contact = *msg;
    g_vmbus->target_vcpu = target_vcpu;
    g_vmbus->int_page = int_page;
    g_vmbus->sint_route = route;
    g_vmbus->state = VMBusState_Initialized;
    return 0;
}

/*
 * Message handling
 */

static void complete_version_response(struct vmbus_channel_message_header *msg, int res)
{
    VMBUS_TRACE("%d", res);
    struct vmbus_channel_version_response *reply =
        (struct vmbus_channel_version_response*) msg;

    /* Cleanup if we failed to send a reply or if we don't like this guest */
    if (res != 0 || reply->status != 0 || !reply->version_supported) {
        error_report("post reply status %d", res);
        hyperv_sint_route_unref(g_vmbus->sint_route);
        unmap_guest_page(g_vmbus->int_page, true);
        return;
    }

    g_vmbus->state = VMBusState_ContactInitiated;
}

/* Process initiate contact message */
static int handle_initiate_contact(
        struct vmbus_channel_initiate_contact *msg,
        uint32_t size)
{
    assert(msg);

    struct vmbus_channel_version_response reply;
    memset(&reply, 0, sizeof(reply));
    reply.header.msgtype = CHANNELMSG_VERSION_RESPONSE;

    if (size < sizeof(*msg)) {
        /* Message sanity check failed - use current vcpu for reply */
        error_report("bad message size %d", size);
        return -EINVAL;
    }

    VMBUS_TRACE("version requested %d, target vcpu %d, "
            "monitor page 1 0x%llx 2 0x%llx, interrupt page 0x%llx",
            msg->vmbus_version_requested,
            msg->target_vcpu,
            (unsigned long long)msg->monitor_page1,
            (unsigned long long)msg->monitor_page2,
            (unsigned long long)msg->interrupt_page);

    qbus_reset_all(&g_vmbus->parent);
    int res = vmbus_init_with_initiate_contact(msg);
    if (res != 0) {
        return res;
    }

    if (msg->vmbus_version_requested <= VMBUS_VERSION_CURRENT) {
        reply.version_supported = true;
    }

    VMBUS_TRACE("sending version response: supported = %d, status = %d",
            reply.version_supported, reply.status);

    return post_reply(&reply.header, sizeof(reply));
}

static int handle_request_offers(void)
{
    /* Find next channel that does not have its offer delivered.
     * We scan full channel list because we don't have a reliable way to save
     * intermidiate position and restore it after incoming migration.
     * After migration device set may be different.
     * Thus channel list order and size may be different as well.
     * Channel list should not be large and offers are handled only during
     * initial handshake stage so it should be fine. */

    VMBusChannel *chan = NULL;
    QLIST_FOREACH(chan, &g_vmbus->channel_list, link) {
        VMBUS_TRACE("channel %p, connid %d", chan, chan->connection_id);
        if (chan->offer_sent) {
            continue;
        }

        struct vmbus_channel_offer_channel reply;
        memset(&reply, 0, sizeof(reply));
        reply.header.msgtype = CHANNELMSG_OFFERCHANNEL;
        reply.child_relid = chan->relid;
        reply.connection_id = chan->connection_id;
        reply.offer = chan->offer;
        chan->offer_sent = true;

        VMBUS_TRACE("channel %p, connid %d", chan, reply.connection_id);
        return post_reply(&reply.header, sizeof(reply));
    }

    /* No more offers */
    struct vmbus_channel_message_header reply;
    memset(&reply, 0, sizeof(reply));
    reply.msgtype = CHANNELMSG_ALLOFFERS_DELIVERED;

    VMBUS_TRACE("alloffers_delivered");
    return post_reply(&reply, sizeof(reply));
}

static void complete_offerchannel(struct vmbus_channel_message_header *msg, int res)
{
    VMBUS_TRACE("%d", res);
    if (res == 0) {
        handle_request_offers();
    }
}

static void complete_alloffers_delivered(struct vmbus_channel_message_header *msg, int res)
{
    VMBUS_TRACE("%d", res);
}

static int reply_gpadl_created(uint32_t gpadl,
        uint32_t relid, uint32_t status)
{
    struct vmbus_channel_gpadl_created reply;
    memset(&reply, 0, sizeof(reply));
    reply.header.msgtype = CHANNELMSG_GPADL_CREATED;
    reply.child_relid = relid;
    reply.gpadl = gpadl;
    reply.creation_status = status;

    return post_reply(&reply.header, sizeof(reply));
}

static void complete_gpadl_created(struct vmbus_channel_message_header *msg, int res)
{
    VMBUS_TRACE("%d", res);
    struct vmbus_channel_gpadl_created *reply = (struct vmbus_channel_gpadl_created*)msg;
    if (res != 0 || reply->creation_status != 0) {
        free_gpadl(find_gpadl(reply->gpadl));
    }
}

static int handle_gpadl_header(
        struct vmbus_channel_gpadl_header *msg,
        uint32_t size)
{
    int status = 0;
    VMBusGpadl *gpadl = NULL;

    assert(msg);

    if (size < sizeof(*msg)) {
        status = -EINVAL;
        goto reply_out;
    }

    if (msg->rangecount != 1) {
        if (msg->rangecount > 1) {
            error_report("gpadl headers with multiple ranges are not supported");
        }

        status = -EINVAL;
        goto reply_out;
    }

    if (find_gpadl(msg->gpadl) != NULL) {
        error_report("gpadl id %d already exists", msg->gpadl);
        status = -ENOENT;
        goto reply_out;
    }

    if ((msg->range[0].byte_count & (TARGET_PAGE_SIZE - 1)) != 0) {
        error_report("bad gpadl size 0x%x", msg->range[0].byte_count);
        status = -EINVAL;
        goto reply_out;
    }

    gpadl = create_gpadl(msg->gpadl, msg->child_relid, msg->range[0].byte_count);

    uint32_t max_pages =
        (size - sizeof(*msg) - sizeof(struct vmbus_gpadl_gpa_range)) / sizeof(uint64_t);
    uint32_t pages = gpadl->total_pages > max_pages ?
        max_pages : gpadl->total_pages;

    uint32_t i;
    for (i = 0; i < pages; ++i) {
        gpadl->gfn[i] = msg->range[0].pfn_array[i];
    }

    gpadl->seen_pages = pages;
    if (gpadl->seen_pages != gpadl->total_pages) {
        /* Not enough PFNs in this message, no reply yet*/
        return 0;
    }

    gpadl->created = true;

reply_out:
    return reply_gpadl_created(msg->gpadl, msg->child_relid, status);
}

static int handle_gpadl_body(struct vmbus_channel_gpadl_body *msg,
        uint32_t size)
{
    int status = 0;
    VMBusGpadl *gpadl = NULL;

    assert(msg);

    if (size < sizeof(*msg)) {
        status = -EINVAL;
        goto reply_out;
    }

    gpadl = find_gpadl(msg->gpadl);
    if (gpadl == NULL) {
        status = -EINVAL;
        goto reply_out;
    }

    uint32_t max_pages = (size - sizeof(*msg)) / sizeof(uint64_t);
    uint32_t pages = gpadl->total_pages - gpadl->seen_pages;
    if (pages > max_pages) {
        pages = max_pages;
    }

    uint32_t i;
    for (i = 0; i < pages; ++i) {
        gpadl->gfn[gpadl->seen_pages + i] = msg->pfn[i];
    }

    gpadl->seen_pages += pages;
    if (gpadl->seen_pages != gpadl->total_pages) {
        /* Not enough PFNs in this message, no reply */
        return 0;
    }

    gpadl->created = true;

reply_out:
    return reply_gpadl_created(msg->gpadl, (gpadl ? gpadl->child_relid : 0),
            status);
}

static void complete_gpadl_torndown(struct vmbus_channel_message_header *msg, int res)
{
    VMBUS_TRACE("%d", res);
    /* Nothing to do */
}

static int handle_gpadl_teardown(
        struct vmbus_channel_gpadl_teardown *msg,
        uint32_t size)
{
    assert(msg);

    if (size < sizeof(*msg)) {
        return -EINVAL;
    }

    VMBusGpadl *gpadl = find_gpadl(msg->gpadl);
    if (!gpadl) {
        /* teardown reply does not include a status field
         * so we can't reply anything and just return here */
        error_report("could not find gpadl id %d", msg->gpadl);
        return -ENOENT;
    }

    if (gpadl->child_relid != msg->child_relid) {
        /* Not sure what to do in this case.
         * We don't need relid argument in this message at all */
        error_report("gpadl is %d has a different relid (%d != %d)",
                msg->gpadl,
                gpadl->child_relid,
                msg->child_relid);
    }

    free_gpadl(gpadl);

    struct vmbus_channel_gpadl_torndown reply;
    memset(&reply, 0, sizeof(reply));
    reply.header.msgtype = CHANNELMSG_GPADL_TORNDOWN;
    reply.gpadl = msg->gpadl;
    return post_reply(&reply.header, sizeof(reply));
}

static void complete_openchannel_result(struct vmbus_channel_message_header *msg, int res)
{
    VMBUS_TRACE("%d", res);
    /* Nothing to do */
}

static int handle_open_channel(
        struct vmbus_channel_open_channel *msg,
        uint32_t size)
{
    assert(msg);

    int status = 0;
    VMBusChannel *chan = NULL;

    struct vmbus_channel_open_result reply;
    memset(&reply, 0, sizeof(reply));

    if (size < sizeof(*msg)) {
        status = -EINVAL;
        goto reply_out;
    }

    chan = find_channel(msg->child_relid);
    if (!chan) {
        status = -EINVAL;
        goto reply_out;
    }

    if (chan->is_open) {
        status = -EINVAL;
        goto reply_out;
    }

    status = open_channel(chan, msg);

reply_out:
    reply.header.msgtype = CHANNELMSG_OPENCHANNEL_RESULT;
    reply.child_relid = msg->child_relid;
    reply.openid = msg->openid;
    reply.status = status;
    return post_reply(&reply.header, sizeof(reply));
}

static int handle_close_channel(
        struct vmbus_channel_close_channel *msg,
        uint32_t size)
{
    assert(msg);

    if (size < sizeof(*msg)) {
        return -EINVAL;
    }

    VMBusChannel *chan = find_channel(msg->child_relid);
    if (!chan) {
        return -ENOENT;
    }

    close_channel(chan);

    /* No reply for this message */
    return 0;
}

static VMBusReplyCompletionHandler g_completion_handlers[] = {
    NULL,
    complete_offerchannel,
    NULL,
    NULL,
    complete_alloffers_delivered,
    NULL,
    complete_openchannel_result,
    NULL,
    NULL,
    NULL,
    complete_gpadl_created,
    NULL,
    complete_gpadl_torndown,
    NULL,
    NULL,
    complete_version_response,
};

static void complete_message(struct vmbus_channel_message_header *msg, int res)
{
    assert(msg);

    if (msg->msgtype >= ARRAY_SIZE(g_completion_handlers) ||
            !g_completion_handlers[msg->msgtype]) {
        error_report("Bad message response type %d", msg->msgtype);
        assert(0);
        return;
    }

    g_completion_handlers[msg->msgtype](msg, res);
}

/* Sint route notification callback */
static void sint_callback(HvSintRoute *route)
{
    VMBUS_TRACE("staged message status %d (%s)",
            route->msg_status,
            strerror(-route->msg_status));

    qemu_bh_schedule(g_vmbus->bh);
}

static bool has_posted_message(void)
{
    return g_vmbus->has_posted_message;
}

/* Post generic reply message to guest on vmbus target vcpu */
static int post_reply(struct vmbus_channel_message_header* reply,
                uint32_t size)
{
    assert(reply);
    assert(g_vmbus->state >= VMBusState_Initialized);

    struct hv_message msg;
    memset(&msg.header, 0, sizeof(msg.header));
    msg.header.message_type = VMBUS_MESSAGE_TYPE;
    msg.header.payload_size = ROUND_UP(size, VMBUS_MESSAGE_PAYLOAD_SIZE_ALIGN);

    if (msg.header.payload_size > sizeof(msg.u.payload)) {
        VMBUS_TRACE("payload size too big %d", msg.header.payload_size);
        return -EINVAL;
    }

    memcpy(msg.u.payload, reply, size);

    VMBUS_TRACE("posting message type %d, size %d",
            reply->msgtype, msg.header.payload_size);

    /* Pre-condition: staging slot is always free on entry */
    assert(g_vmbus->sint_route->msg_status == 0);
    assert(!has_posted_message());

    g_vmbus->has_posted_message = true;
    hyperv_post_msg(g_vmbus->sint_route, &msg);
    return 0;
}

/* Handles any messages that are currently being posted to guest */
static void process_posted_message(void)
{
    if (!has_posted_message()) {
        return;
    }

    struct vmbus_channel_message_header *hdr =
        (struct vmbus_channel_message_header*)g_vmbus->sint_route->msg.u.payload;

    switch (g_vmbus->sint_route->msg_status) {
        case -EAGAIN:   /* Repeat same message */
            g_vmbus->sint_route->msg_status = 0;
            hyperv_post_msg(g_vmbus->sint_route, &g_vmbus->sint_route->msg);
        case -EBUSY:    /* Ignore */
            break;
        case 0:         /* Message delivered, fallthrough */
        default:        /* System error */
            g_vmbus->has_posted_message = false;
            complete_message(hdr, g_vmbus->sint_route->msg_status);
            break;
    }
}

/* Handle an incoming message */
static void process_incoming_message(void *opaque)
{
    int res = 0;
    struct hv_input_post_message *msg = (struct hv_input_post_message*)opaque;

    uint32_t payload_size = msg->payload_size;
    struct vmbus_channel_message_header *vmbus_msg =
        (struct vmbus_channel_message_header*)msg->payload;

    /* Check valid vmbus state for this message, drop invalid messages */
    if (vmbus_msg->msgtype > CHANNELMSG_INITIATE_CONTACT) {
        if (g_vmbus->state < VMBusState_ContactInitiated) {
            error_report("Invalid state %d for message type %d",
                    g_vmbus->state, vmbus_msg->msgtype);
            return;
        }
    }

    /* Handle message type */
    switch (vmbus_msg->msgtype) {

        case CHANNELMSG_INITIATE_CONTACT:
            res = handle_initiate_contact(
                (struct vmbus_channel_initiate_contact*)vmbus_msg,
                payload_size);
            break;

        case CHANNELMSG_REQUESTOFFERS:
            res = handle_request_offers();
            break;

        case CHANNELMSG_GPADL_HEADER:
            res = handle_gpadl_header(
                    (struct vmbus_channel_gpadl_header*)vmbus_msg,
                    payload_size);
            break;

        case CHANNELMSG_GPADL_BODY:
            res = handle_gpadl_body(
                    (struct vmbus_channel_gpadl_body*)vmbus_msg,
                    payload_size);
            break;

        case CHANNELMSG_GPADL_TEARDOWN:
            res = handle_gpadl_teardown(
                    (struct vmbus_channel_gpadl_teardown*)vmbus_msg,
                    payload_size);
            break;

        case CHANNELMSG_OPENCHANNEL:
            res = handle_open_channel(
                    (struct vmbus_channel_open_channel*)vmbus_msg,
                    payload_size);
            break;

        case CHANNELMSG_CLOSECHANNEL:
            res = handle_close_channel(
                    (struct vmbus_channel_close_channel*)vmbus_msg,
                    payload_size);
            break;

        default:
            error_report("unknown message type 0x%x", vmbus_msg->msgtype);
            res = -EINVAL;
            break;
    }

    if (res != 0) {
        /* VMBus does not define specific packet type to report async errors
         * Real HyperV usually just doesn't reply when it sees a bad packet */
        error_report("%s: failed with status %d", __FUNCTION__, res);
    }
}

/* VMBus state machine function */
static void bh_worker(void* opaque)
{
    VMBUS_TRACE("enter");

    /* Handle pending reply message if any */
    process_posted_message();

    /* Handle incoming message if we've dealt with posted reply */
    struct hv_input_post_message msg;
    if (!has_posted_message() && dequeue_incoming_message(&msg)) {
        process_incoming_message(&msg);

        /* Incoming message handler might have posted a new message, handle that now */
        process_posted_message();

        /* Reschedule bh if we have more incoming messages */
        if (have_incoming_message()) {
            qemu_bh_schedule(g_vmbus->bh);
        }
    }
}

/*
 * Called from hyperv hypercall handler to process vmbus post message call
 * Executes in VCPU thread, possibly concurrently with other VCPU threads
 */
int vmbus_hcall_post_message(X86CPU *vcpu, struct kvm_hyperv_exit *exit)
{
    if (!vcpu || !exit) {
        return HV_STATUS_INVALID_HYPERCALL_INPUT;
    }

    if (hyperv_is_fast_hcall(exit)) {
        return HV_STATUS_INVALID_HYPERCALL_INPUT;
    }

    uint64_t ingpa = exit->u.hcall.params[0];
    void *post_msg_page = map_guest_page(ingpa, false);
    if (!post_msg_page) {
        return HV_STATUS_INSUFFICIENT_MEMORY;
    }

    struct hv_input_post_message post_msg;
    memcpy(&post_msg, (struct hv_input_post_message*)post_msg_page,
            sizeof(post_msg));

    unmap_guest_page(post_msg_page, false);

    union hv_connection_id conn_id = post_msg.connectionid;
    uint32_t message_type = post_msg.message_type;
    uint32_t payload_size = post_msg.payload_size;

    if (conn_id.u.id != VMBUS_MESSAGE_CONNECTION_ID) {
        return HV_STATUS_INVALID_CONNECTION_ID;
    }
    if (message_type != VMBUS_MESSAGE_TYPE) {
        return HV_STATUS_INVALID_HYPERCALL_INPUT;
    }
    if (payload_size > sizeof(post_msg.payload)) {
        return HV_STATUS_INVALID_HYPERCALL_INPUT;
    }
    if (payload_size < sizeof(struct vmbus_channel_message_header)) {
        return HV_STATUS_INVALID_HYPERCALL_INPUT;
    }

    struct vmbus_channel_message_header *vmbus_msg =
        (struct vmbus_channel_message_header*)post_msg.payload;

    VMBUS_TRACE("message type %d, size %d", vmbus_msg->msgtype, payload_size);

    if (vmbus_msg->msgtype == CHANNELMSG_INVALID ||
        vmbus_msg->msgtype >= CHANNELMSG_COUNT) {
        error_report("vmbus: unknown message type 0x%x", vmbus_msg->msgtype);
        return HV_STATUS_INVALID_HYPERCALL_INPUT;
    }

    if (!enqueue_incoming_message(&post_msg)) {
        error_report("post message queue is full");
        return HV_STATUS_INSUFFICIENT_BUFFERS;
    }

    return HV_STATUS_SUCCESS;
}

/**
 * Handles HVCALL_SIGNAL_EVENT hypercall
 * Executes in VCPU thread
 */
int vmbus_hcall_signal_event(X86CPU *vcpu, struct kvm_hyperv_exit *exit)
{
    if (!vcpu || !exit) {
        return HV_STATUS_INVALID_HYPERCALL_INPUT;
    }

    uint32_t connid = 0;
    if (!hyperv_is_fast_hcall(exit)) {
        void *msg_page = map_guest_page(exit->u.hcall.params[0], false);
        if (!msg_page) {
            return HV_STATUS_INSUFFICIENT_MEMORY;
        }

        struct hv_input_signal_event *input = (struct hv_input_signal_event *)msg_page;
        connid = input->connectionid.u.id;
        unmap_guest_page(msg_page, false);
    } else {
        connid = exit->u.hcall.params[0];
    }

    if (connid == VMBUS_EVENT_CONNECTION_ID) {
        /* Guest uses interrupt page which may have several events set */
        struct hv_interrupt_page *int_page = g_vmbus->int_page;
        if (!int_page) {
            return HV_STATUS_INVALID_PARAMETER;
        }

        hv_status status = HV_STATUS_SUCCESS;
        unsigned i;
        for (i = 0; i < ARRAY_SIZE(int_page->from_guest); ++i) {
            if (int_page->from_guest[i] == 0) {
                continue;
            }

            unsigned bit;
            for (bit = 0; bit < BITS_PER_LONG; ++bit) {
                unsigned long mask = 1ul << bit;
                if (atomic_fetch_and(&int_page->from_guest[i], ~mask) != mask) {
                    continue;
                }

                uint32_t relid = i * BITS_PER_LONG + bit;
                VMBusChannel *chan = find_channel(relid);
                if (!chan) {
                    /* Remember error and continue */
                    error_report("no channel for relid %d", relid);
                    status = HV_STATUS_INVALID_PARAMETER;
                    continue;
                }

                event_notifier_set(&chan->owner->signal_event);
            }
        }

        return status;
    } else {
        /* Use channel connection id directly */
        VMBusChannel *chan = find_channel_connid(connid);
        if (!chan) {
            error_report("no channel for connid %d", connid);
            return HV_STATUS_INVALID_PARAMETER;
        }

        event_notifier_set(&chan->owner->signal_event);
        return HV_STATUS_SUCCESS;
    }

    assert(0 && "Unreachable");
}

VMBusChannel* vmbus_create_channel(VMBusDevice *dev, uint16_t chn_flags)
{
    QemuUUID tmp_uuid;

    if (g_vmbus->state >= VMBusState_OffersDelivered) {
        return NULL;
    }

    VMBusChannel *chan = alloc_channel();

    /* Hyper-V wants LE GUIDs */
    tmp_uuid = VMBUS_DEVICE_GET_CLASS(dev)->class_uuid;
    qemu_uuid_bswap(&tmp_uuid);
    memcpy(chan->offer.if_type.b, tmp_uuid.data, sizeof(tmp_uuid.data));

    tmp_uuid = dev->instanceid;
    qemu_uuid_bswap(&tmp_uuid);
    memcpy(chan->offer.if_instance.b, tmp_uuid.data, sizeof(tmp_uuid.data));

    chan->offer.chn_flags = chn_flags;

    chan->owner = dev;
    chan->relid = alloc_chan_relid();
    chan->connection_id = alloc_connection_id();
    return chan;
}

void vmbus_release_channel(VMBusChannel *chan)
{
    if (chan) {
        free_connection_id(chan->connection_id);
        free_chan_relid(chan->relid);
        free_channel(chan);
    }
}

int vmbus_channel_send(struct VMBusChannel *chan,
        const void *buf, uint32_t buflen, enum vmbus_packet_type type,
        uint64_t request_id)
{
    int err = 0;
    struct vmpacket_descriptor desc;
    uint32_t packetlen = sizeof(struct vmpacket_descriptor) + buflen;
    uint32_t packetlen_aligned = ROUND_UP(packetlen, sizeof(uint64_t));
    struct hv_kvec bufferlist[3];
    uint64_t aligned_data = 0;

    if (!chan || !chan->is_open) {
        return -EINVAL;
    }

    desc.type = type;
    desc.flags = 0;
    desc.offset8 = sizeof(struct vmpacket_descriptor) >> 3;
    desc.len8 = (uint16_t)(packetlen_aligned >> 3);
    desc.trans_id = request_id;

    bufferlist[0].iov_base = &desc;
    bufferlist[0].iov_len = sizeof(struct vmpacket_descriptor);
    bufferlist[1].iov_base = (void*)buf;
    bufferlist[1].iov_len = buflen;
    bufferlist[2].iov_base = &aligned_data;
    bufferlist[2].iov_len = (packetlen_aligned - packetlen);

    bool signal = false;
    err = hv_ringbuffer_write(&chan->rb_outbound, bufferlist,
            ARRAY_SIZE(bufferlist), &signal);
    if (err != 0) {
        error_report("hv_ringbuffer_write failed with %d", err);
    } else if (signal) {
        err = signal_channel(chan);
        if (err != 0) {
            return err;
        }
    }

    return 0;
}

int vmbus_channel_peek(struct VMBusChannel *chan, struct vmpacket_descriptor *out_desc)
{
    if (!chan || !chan->is_open) {
        return -EINVAL;
    }

    if (!out_desc) {
        return -EINVAL;
    }

    return hv_ringbuffer_peek(&chan->rb_inbound, out_desc, sizeof(*out_desc));
}

int vmbus_channel_recv(struct VMBusChannel *chan, void *buf, uint32_t buflen,
        uint32_t *msglen, uint32_t flags)
{
    int ret = 0;
    struct vmpacket_descriptor desc;

    if (!chan || !chan->is_open) {
        return -EINVAL;
    }

    ret = hv_ringbuffer_peek(&chan->rb_inbound, &desc, sizeof(desc));
    if (ret != 0) {
        return ret;
    }

    uint32_t len = vmpacket_get_length(&desc);
    uint32_t offset = vmpacket_get_offset(&desc);

    if (len < offset) {
        return -EINVAL;
    }

    uint32_t dataoff = (flags & VMBusChannelRecv_Raw ? 0 : offset);
    uint32_t datalen = len - dataoff;

    VMBUS_TRACE("pending message type %d, len %d, offset %d, flags 0x%x, request id 0x%llx",
            desc.type, datalen, offset, desc.flags, (unsigned long long)desc.trans_id);

    if (msglen) {
        *msglen = datalen;
    }

    if (buflen < datalen) {
        return -ENOSPC;
    }

    if (buf) {
        bool signal = false;
        ret = hv_ringbuffer_read(&chan->rb_inbound, buf, datalen, dataoff,
                &signal);
        if (ret == 0 && signal) {
            ret = signal_channel(chan);
            if (ret != 0) {
                return ret;
            }
        }
    }

    return ret;
}

/*
 * Migration
 */

/* struct hv_input_post_message */
static const VMStateDescription vmbus_vmstate_input_post_message = {
    .name = "vmbus/hv_input_post_message",
    .version_id = VMBUS_VMSTATE_VERSION_ID,
    .minimum_version_id = VMBUS_VMSTATE_VERSION_ID,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32(connectionid.asu32, struct hv_input_post_message),
        VMSTATE_UINT32(message_type, struct hv_input_post_message),
        VMSTATE_UINT32(payload_size, struct hv_input_post_message),
        VMSTATE_UINT64_ARRAY(payload, struct hv_input_post_message, HV_MESSAGE_PAYLOAD_QWORD_COUNT),
        VMSTATE_END_OF_LIST()
    }
};

/* struct hv_message */
static const VMStateDescription vmbus_vmstate_hv_message = {
    .name = "vmbus/hv_message",
    .version_id = VMBUS_VMSTATE_VERSION_ID,
    .minimum_version_id = VMBUS_VMSTATE_VERSION_ID,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32(header.message_type, struct hv_message),
        VMSTATE_UINT8(header.payload_size, struct hv_message),
        VMSTATE_UINT8(header.message_flags.asu8, struct hv_message),
        VMSTATE_UINT64(header.sender, struct hv_message),
        VMSTATE_UINT64_ARRAY(u.payload, struct hv_message, HV_MESSAGE_PAYLOAD_QWORD_COUNT),
        VMSTATE_END_OF_LIST()
    }
};

/* VMBus generic device */
static const VMStateDescription vmbus_device_state_description = {
    .name = TYPE_VMBUS_DEVICE,
    .version_id = VMBUS_VMSTATE_VERSION_ID,
    .minimum_version_id = VMBUS_VMSTATE_VERSION_ID,
    .fields = (VMStateField[]) {
        VMSTATE_UINT8_ARRAY(instanceid.data, VMBusDevice, 16),
        VMSTATE_STRUCT_POINTER(chan, VMBusDevice, vmbus_vmstate_channel, VMBusChannel),
        VMSTATE_END_OF_LIST()
    }
};

static void vmbus_save(QEMUFile *f, void *opaque, size_t size)
{
    size_t i = 0;
    uint32_t total = 0;

    /*
     * General state
     */
    qemu_put_be32(f, g_vmbus->state);
    qemu_put_buffer(f, (uint8_t*)&g_vmbus->initiate_contact, sizeof(g_vmbus->initiate_contact));
    if (g_vmbus->state >= VMBusState_Initialized) {
        qemu_put_be32(f, g_vmbus->sint_route->msg_status);
        vmstate_save_state(f, &vmbus_vmstate_hv_message, &g_vmbus->sint_route->msg, NULL);
    }

    qemu_put_buffer(f, (uint8_t*)g_vmbus->chan_relid_bitmap, sizeof(g_vmbus->chan_relid_bitmap));
    qemu_put_buffer(f, (uint8_t*)g_vmbus->connection_id_bitmap, sizeof(g_vmbus->connection_id_bitmap));

    /*
     * RX queue state
     *
     * rx_queue_lock is ignored
     * Lock protects concurrent VCPU access to rx_queue and there is no VCPU
     * threads running during migration so lock is always free
     */
    qemu_put_be64(f, g_vmbus->rx_queue_head);
    qemu_put_be64(f, g_vmbus->rx_queue_size);
    for (i = 0; i < g_vmbus->rx_queue_size; ++i) {
        vmstate_save_state(f, &vmbus_vmstate_input_post_message,
                &g_vmbus->rx_queue[get_queue_index(i)], NULL);
    }

    /*
     * gpadl list
     */
    VMBusGpadl *gpadl = NULL;
    total = 0;
    QLIST_FOREACH(gpadl, &g_vmbus->gpadl_list, link) {
        ++total;
    }
    qemu_put_be32(f, total);
    QLIST_FOREACH(gpadl, &g_vmbus->gpadl_list, link) {
        save_gpadl(f, gpadl);
    }
}

static int vmbus_load(QEMUFile *f, void *opaque, size_t size)
{
    int ret = 0;
    uint32_t i = 0;

    assert(g_vmbus->state == VMBusState_Reset);

    uint32_t state = qemu_get_be32(f);
    struct vmbus_channel_initiate_contact msg;
    qemu_get_buffer(f, (uint8_t*)&msg, sizeof(msg));
    if (state >= VMBusState_Initialized) {
        ret = vmbus_init_with_initiate_contact(&msg);
        if (ret != 0) {
            return ret;
        }

        g_vmbus->state = state;
        g_vmbus->sint_route->msg_status = qemu_get_be32(f);
        ret = vmstate_load_state(f, &vmbus_vmstate_hv_message,
                &g_vmbus->sint_route->msg, VMBUS_VMSTATE_VERSION_ID);
        if (ret != 0) {
            return ret;
        }
    }

    qemu_get_buffer(f, (uint8_t*)g_vmbus->chan_relid_bitmap, sizeof(g_vmbus->chan_relid_bitmap));
    qemu_get_buffer(f, (uint8_t*)g_vmbus->connection_id_bitmap, sizeof(g_vmbus->connection_id_bitmap));

    g_vmbus->rx_queue_head = qemu_get_be64(f);
    g_vmbus->rx_queue_size = qemu_get_be64(f);
    for (i = 0; i < g_vmbus->rx_queue_size; ++i) {
        ret = vmstate_load_state(f, &vmbus_vmstate_input_post_message,
                &g_vmbus->rx_queue[get_queue_index(i)], VMBUS_VMSTATE_VERSION_ID);
        if (ret != 0) {
            return ret;
        }
    }

    uint32_t total_gpadls = qemu_get_be32(f);
    for (i = 0; i < total_gpadls; ++i) {
        ret = load_gpadl(f);
        if (ret != 0) {
            return ret;
        }
    }

    /* Resume processing */
    qemu_bh_schedule(g_vmbus->bh);
    return 0;
}

static const VMStateInfo vmbus_vmstate_info = {
    .name = "vmbus",
    .get = vmbus_load,
    .put = vmbus_save,
};

static const VMStateDescription vmbus_vmstate_description = {
    .name = TYPE_VMBUS_ROOT_DEV,
    .version_id = 1,
    .minimum_version_id = 1,
    .fields = (VMStateField[]) {
        {
            .name = "vmbus",
            .info = &vmbus_vmstate_info,
            .flags = VMS_SINGLE,
        },
        VMSTATE_END_OF_LIST()
    }
};

/*
 * qdev stuff
 */

static void vmbus_dev_realize(DeviceState *dev, Error **errp)
{
    VMBUS_TRACE("%p", dev);

    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusDeviceClass *vdc = VMBUS_DEVICE_GET_CLASS(vdev);
    Error *err = NULL;

    if (vdc->singleton && vdc->instance_count > 0) {
        error_setg(&err, "trying to create instance of a singleton device");
        goto error_out;
    }

    if (qemu_uuid_is_null(&vdev->instanceid) && !vdc->singleton) {
        qemu_uuid_generate(&vdev->instanceid);
    }

    if (vdc->realize) {
        vdc->realize(dev, &err);
        if (err != NULL) {
            goto error_out;
        }
    }

    int res = vmstate_register(dev, -1, &vmbus_device_state_description, dev);
    if (res != 0) {
        error_setg(&err, "vmstate_register says %d", res);
        goto error_out;
    }

    vdc->instance_count++;
    return;

error_out:
    error_propagate(errp, err);
}

static void vmbus_dev_reset(DeviceState *dev)
{
    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusDeviceClass *vdc = VMBUS_DEVICE_GET_CLASS(vdev);

    VMBUS_TRACE("%p", dev);

    if (vdev->chan) {
        close_channel(vdev->chan);
    }

    if (vdc->reset) {
        vdc->reset(dev);
    }
}

static void vmbus_dev_unrealize(DeviceState *dev, Error **errp)
{
    VMBUS_TRACE("%p", dev);

    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    VMBusDeviceClass *vdc = VMBUS_DEVICE_GET_CLASS(vdev);
    Error *err = NULL;

    if (vdc->unrealize) {
        vdc->unrealize(dev, &err);
        if (err != NULL) {
            error_propagate(errp, err);
            return;
        }
    }

    free_channel(vdev->chan);
}

static Property vmbus_dev_properties[] = {
    DEFINE_PROP_UUID("instanceid", VMBusDevice, instanceid),
    DEFINE_PROP_END_OF_LIST(),
};

static void vmbus_dev_class_init(ObjectClass *klass, void *data)
{
    VMBUS_TRACE("%p", klass);
    DeviceClass *kdev = DEVICE_CLASS(klass);
    kdev->props = vmbus_dev_properties;
    kdev->bus_type = TYPE_VMBUS;
    kdev->realize = vmbus_dev_realize;
    kdev->unrealize = vmbus_dev_unrealize;
    kdev->reset = vmbus_dev_reset;
}

/* vmbus generic device base */
static const TypeInfo vmbus_dev_type_info = {
    .name = TYPE_VMBUS_DEVICE,
    .parent = TYPE_DEVICE,
    .abstract = true,
    .instance_size = sizeof(VMBusDevice),
    .class_size = sizeof(VMBusDeviceClass),
    .class_init = vmbus_dev_class_init,
};

static void vmbus_realize(BusState *bus, Error **errp)
{
    VMBus *vmbus = VMBUS_STATE(bus);
    VMBUS_TRACE("%p", bus);

    qemu_mutex_init(&vmbus->rx_queue_lock);
    vmbus->bh = qemu_bh_new(bh_worker, NULL);

    QLIST_INIT(&vmbus->gpadl_list);
    QLIST_INIT(&vmbus->channel_list);
}

static void vmbus_unrealize(BusState *bus, Error **errp)
{
    VMBus *vmbus = VMBUS_STATE(bus);
    VMBUS_TRACE("%p", bus);

    qemu_bh_delete(vmbus->bh);
    qemu_mutex_destroy(&vmbus->rx_queue_lock);
}

static void vmbus_reset(BusState *bus)
{
    VMBus *vmbus = VMBUS_STATE(bus);
    VMBUS_TRACE("%p", bus);

    if (vmbus->state == VMBusState_Reset) {
        VMBUS_TRACE("No reset is nessesary");
        return;
    }

    /* Free any remaining gpadl's */
    while (!QLIST_EMPTY(&vmbus->gpadl_list)) {
        VMBusGpadl *gpadl = QLIST_FIRST(&vmbus->gpadl_list);
        free_gpadl(gpadl);
    }

    memset(&vmbus->initiate_contact, 0, sizeof(vmbus->initiate_contact));
    hyperv_sint_route_unref(g_vmbus->sint_route);
    unmap_guest_page(vmbus->int_page, true);
    vmbus->sint_route = NULL;
    vmbus->int_page = NULL;
    vmbus->target_vcpu = NULL;

    vmbus->state = VMBusState_Reset;
    VMBUS_TRACE("Reset done");
}

static char *vmbus_get_dev_path(DeviceState *dev)
{
    BusState *bus = qdev_get_parent_bus(dev);
    DeviceState *proxy = DEVICE(bus->parent);
    return qdev_get_dev_path(proxy);
}

static char *vmbus_get_fw_dev_path(DeviceState *dev)
{
    DeviceClass *vdc = DEVICE_GET_CLASS(dev);
    VMBusDevice *vdev = VMBUS_DEVICE(dev);
    char path[32 + 1 + UUID_FMT_LEN + 1];
    int off;

    off = snprintf(path, sizeof(path), "%s@", vdc->fw_name);
    if (sizeof(path) - off > UUID_FMT_LEN) {
        qemu_uuid_unparse(&vdev->instanceid, path + off);
    }
    return g_strdup(path);
}

static void vmbus_bus_class_init(ObjectClass *klass, void *data)
{
    BusClass *k = BUS_CLASS(klass);

    k->get_dev_path = vmbus_get_dev_path;
    k->get_fw_dev_path = vmbus_get_fw_dev_path;
    k->realize = vmbus_realize;
    k->unrealize = vmbus_unrealize;
    k->reset = vmbus_reset;
}

/* vmbus bus type */
static const TypeInfo vmbus_bus_type_info = {
    .name = TYPE_VMBUS,
    .parent = TYPE_BUS,
    .instance_size = sizeof(VMBus),
    .class_init = vmbus_bus_class_init,
};

static void vmbus_root_dev_realize(DeviceState *dev, Error **errp)
{
    /* There should be only 1 instace of vmbus root device */
    if (NULL != qdev_find_recursive(sysbus_get_default(), TYPE_VMBUS_ROOT_DEV)) {
        Error *local_err = NULL;
        error_setg(&local_err, "vmbus root device already exists");
        error_propagate(errp, local_err);
        return;
    }

    g_vmbus = VMBUS_STATE(qbus_create(TYPE_VMBUS, dev, VMBUS_NAME));

    int res = vmstate_register(dev, -1, &vmbus_vmstate_description, dev);
    if (res != 0) {
        error_report("vmstate_register says %d", res);
        exit(0);
    }
}

static void vmbus_root_dev_class_init(ObjectClass *klass, void *data)
{
    VMBUS_TRACE("enter");
    DeviceClass *k = DEVICE_CLASS(klass);

    k->realize = vmbus_root_dev_realize;
    k->fw_name = "vmbus";
    k->cannot_instantiate_with_device_add_yet = true;
}

/* Pseudo-device to connect vmbus to qemu system bus */
static const TypeInfo vmbus_root_dev_type_info = {
    .name = TYPE_VMBUS_ROOT_DEV,
    .parent = TYPE_SYS_BUS_DEVICE,
    .instance_size = sizeof(SysBusDevice),
    .class_init = vmbus_root_dev_class_init,
};

void vmbus_create(void)
{
    DeviceState *dev = qdev_create(NULL, TYPE_VMBUS_ROOT_DEV);
    qdev_init_nofail(dev);
}

bool vmbus_exists(void)
{
    return g_vmbus != NULL;
}

static void vmbus_root_dev_register_types(void)
{
    type_register_static(&vmbus_root_dev_type_info);
    type_register_static(&vmbus_dev_type_info);
    type_register_static(&vmbus_bus_type_info);
}

type_init(vmbus_root_dev_register_types)
