/*
 * QEMU Hyper-V VMBus support
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 */

#ifndef QEMU_VMBUS_H
#define QEMU_VMBUS_H

#include "hw/qdev.h"
#include "sysemu/sysemu.h"
#include "target-i386/hyperv.h"
#include "target-i386/hyperv_defs.h"
#include "hw/vmbus/ring_buffer.h"
#include "hw/vmbus/pagemap.h"
#include "qemu/uuid.h"

//#define VMBUS_DEBUG

#if defined(VMBUS_DEBUG)
#   include "qemu/error-report.h"
#   define VMBUS_TRACE(fmt, ...) \
        error_report("%s: " fmt, __FUNCTION__, ##__VA_ARGS__)
#else
#   define VMBUS_TRACE(fmt, ...)
#endif

#define VMBUS_NAME "vmbus"

#define TYPE_VMBUS "vmbus"
#define TYPE_VMBUS_ROOT_DEV "vmbus-root-dev"
#define TYPE_VMBUS_DEVICE "vmbus-dev"

#define VMBUS_DEVICE(obj) \
    OBJECT_CHECK(VMBusDevice, (obj), TYPE_VMBUS_DEVICE)
#define VMBUS_DEVICE_CLASS(klass) \
    OBJECT_CLASS_CHECK(VMBusDeviceClass, (klass), TYPE_VMBUS_DEVICE)
#define VMBUS_DEVICE_GET_CLASS(obj) \
    OBJECT_GET_CLASS(VMBusDeviceClass, (obj), TYPE_VMBUS_DEVICE)

#define TYPE_VMBUS_HEARTBEAT "hv-heartbeat"
#define VMBUS_HEARTBEAT(obj) \
        OBJECT_CHECK(VMBusHeartbeat, (obj), TYPE_VMBUS_HEARTBEAT)

#define TYPE_VMBUS_SCSI "hv-scsi"
#define VMBUS_SCSI(obj) \
        OBJECT_CHECK(VMBusScsi, (obj), TYPE_VMBUS_SCSI)

struct VMBusChannel;
struct VMBusDevice;

/**
 * Base device class for all devices on VMBus
 */
typedef struct VMBusDeviceClass {
    DeviceClass parent;

    /* Only a fixed device instance is permitted, with a fixed instance uuid */
    bool singleton;
    int instance_count;
    QemuUUID class_uuid;
    DeviceRealize realize;
    DeviceUnrealize unrealize;
    void (*reset)(DeviceState *dev);
    int (*open_channel) (struct VMBusDevice *vdev, unsigned char *userdata);
    void (*close_channel) (struct VMBusDevice *vdev);
} VMBusDeviceClass;

/**
 * Base device class instance
 */
typedef struct VMBusDevice {
    DeviceState parent;
    struct VMBusChannel *chan;
    EventNotifier signal_event;
    QemuUUID instanceid;
} VMBusDevice;

/*
 * Tracks GPADL created by guest request
 * GPADL is basically a ringbuffer on shred guest/host memory
 */
typedef struct VMBusGpadl
{
    QLIST_ENTRY(VMBusGpadl) link;

    uint32_t id;            /* Guest-supplied id */
    uint32_t child_relid;   /* Channel id this gpadl is for */
    bool created;           /* Gpadl created, same as total == seen pages */

    uint32_t total_pages;   /* Total GFN pages */
    uint32_t seen_pages;    /* GFN pages reported by guest so far */
    uint64_t gfn[0];        /* GFN array, total_pages in size */
} VMBusGpadl;

/*
 * vmbus channel is an established communication context between
 * a specific device's server side and guest driver side
 */
typedef struct VMBusChannel
{
    VMBusDevice *owner;         /* Device that owns this channel */

    /* Channel offer message, should be initialized by device */
    struct vmbus_channel_offer offer;
    bool offer_sent;

    uint32_t relid;             /* Channel id */
    uint32_t connection_id;     /* Tracked connection id, used in host signalling */

    bool is_open;               /* Fields below only valid if channel is open */

    uint32_t target_vcpu;       /* Target vcpu to deliver events on */
    HvSintRoute *notify_route;  /* Sint route used to signal events on this channel */

    uint32_t rb_gpadl;          /* Ringbuffer gpadl */
    uint32_t rb_page_offset;    /* Ringbuffer downstream page offset */

    struct hv_ring_buffer_info rb_inbound;  /* Inbound part of ringbuffer */
    struct hv_ring_buffer_info rb_outbound; /* Outbound part of ringbuffer */

    QLIST_ENTRY(VMBusChannel) link;
} VMBusChannel;

/**
 * Create global vmbus instance if needed
 */
void vmbus_create(void);

/**
 * Tell if vmbus is created in realized
 */
bool vmbus_exists(void);

/**
 * Process post message hypercall
 */
int vmbus_hcall_post_message(X86CPU *vcpu, struct kvm_hyperv_exit *exit);

/**
 * Process signal event hypercall
 */
int vmbus_hcall_signal_event(X86CPU *vcpu, struct kvm_hyperv_exit *exit);

/**
 * Create device channel
 * Should be called by vmbus devices on initialization
 * All channel offers are sent over when guest requests them
 */
VMBusChannel* vmbus_create_channel(VMBusDevice *dev, uint16_t chn_flags);
void vmbus_release_channel(VMBusChannel *chan);

/**
 * Send a packet through vmbus channel
 *
 * @chan        Channel to use
 * @buf         Message data buffer
 * @buflen      Message data length in bytes
 * @type        Packet type
 * @request_id  Optional request id to attach to packet
 */
int vmbus_channel_send(struct VMBusChannel *chan,
        const void *buf, uint32_t buflen, enum vmbus_packet_type type,
        uint64_t request_id);

/** Flags for vmbus_channel_recv */
typedef enum VMBusChannelRecvFlags {
    VMBusChannelRecv_Raw = 0x0001,  /** Recv raw packet buffer (including descriptor)
                                      otherwise will try to recv only packet data */
} VMBusChannelRecvFlags;

/**
 * Recv a packet through vmbus channel without blocking
 *
 * @chan        Channel to use
 * @buf         Message output buffer.
 *              Can be NULL if caller wants only pending message size
 * @buflen      Allocated buffer size in bytes
 * @recv_len    Actual received message size.
 *              If buffer is NULL will report pending message size here
 * @flags       See VMBusChannelRecvFlags
 *
 * @returns     -ENOSPC if buffer size is not enough to hold pending message
 *              required buffer size is returned in recv_len
 */
int vmbus_channel_recv(struct VMBusChannel *chan, void *buf, uint32_t buflen,
        uint32_t *recv_len, uint32_t flags);

/** A very common case of channel recv when caller is only interested in message data */
static inline int vmbus_channel_recv_data(struct VMBusChannel *chan, void *buf,
        uint32_t buflen, uint32_t *recvlen)
{
    return vmbus_channel_recv(chan, buf, buflen, recvlen, 0);
}

static inline int vmbus_channel_recv_raw(struct VMBusChannel *chan, void *buf,
        uint32_t buflen, uint32_t *recvlen)
{
    return vmbus_channel_recv(chan, buf, buflen, recvlen, VMBusChannelRecv_Raw);
}

/**
 * Read incoming message descriptor.
 * Calling this function will not move channel ring buffer pointer
 *
 * @chan        Channel to use
 * @out_desc    Returned fixed size message descriptor
 */
int vmbus_channel_peek(struct VMBusChannel *chan, struct vmpacket_descriptor *out_desc);

#endif
