/*
 * QEMU KVM Hyper-V support
 *
 * Copyright (C) 2015 Andrey Smetanin <asmetanin@virtuozzo.com>
 *
 * Authors:
 *  Andrey Smetanin <asmetanin@virtuozzo.com>
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 *
 */

#include "qemu/osdep.h"
#include "qemu/main-loop.h"
#include "qemu/bitops.h"
#include "qapi/error.h"
#include "qapi/visitor.h"
#include "hyperv.h"
#include "standard-headers/asm-x86/hyperv.h"
#include "hyperv_defs.h"
#include "qemu/main-loop.h"
#include "hw/vmbus/vmbus.h"

typedef struct SynICState {
    /*< private >*/
    Object parent_obj;
    /*< public >*/
} SynICState;
#define TYPE_SYNIC "hyperv-synic"
#define SYNIC(obj) OBJECT_CHECK(SynICState, (obj), TYPE_SYNIC)

typedef struct SINTState {
    /*< private >*/
    Object parent_obj;
    /*< public >*/
    uint8_t sint_no;
} SINTState ;
#define TYPE_SINT "hyperv-sint"
#define SINT(obj) OBJECT_CHECK(SINTState, (obj), TYPE_SINT)

#define TYPE_SINT_ROUTE "hyperv-sint-route"
#define SINT_ROUTE(obj) OBJECT_CHECK(HvSintRoute, (obj), TYPE_SINT_ROUTE)

X86CPU *hyperv_find_vcpu(uint32_t vp_index)
{
    CPUState *cpu;
    CPU_FOREACH(cpu) {
        if (cpu->cpu_index == vp_index) {
            return X86_CPU(cpu);
        }
    }
    return NULL;
}

static hwaddr synic_msg_page_addr(X86CPU *cpu)
{
    CPUX86State *env = &cpu->env;
    if (!(env->msr_hv_synic_control & HV_SYNIC_CONTROL_ENABLE) ||
        !(env->msr_hv_synic_msg_page & HV_SYNIC_SIMP_ENABLE)) {
        return 0;
    }
    return env->msr_hv_synic_msg_page & TARGET_PAGE_MASK;
}

static hwaddr synic_evt_page_addr(X86CPU *cpu)
{
    CPUX86State *env = &cpu->env;
    if (!(env->msr_hv_synic_control & HV_SYNIC_CONTROL_ENABLE) ||
        !(env->msr_hv_synic_evt_page & HV_SYNIC_SIEFP_ENABLE)) {
        return 0;
    }
    return env->msr_hv_synic_evt_page & TARGET_PAGE_MASK;
}

/*
 * Worker to transfer the message the from staging area into the guest-owned
 * message page in vcpu context, which guarantees serialization with both KVM
 * vcpu and the guest cpu.  OTOH it may run concurrently with the message
 * originator so care must be taken accessing the staged_msg.
 */
static void cpu_post_msg(void* data)
{
    int ret;
    HvSintRoute *sint_route = (HvSintRoute*)data;
    X86CPU *cpu = sint_route->cpu;
    struct hv_message_page *msg_page;
    struct hv_message *dst_msg;
    hwaddr msg_page_addr;
    hwaddr len = 1;

    msg_page_addr = synic_msg_page_addr(cpu);
    if (!msg_page_addr) {
        ret = -ENXIO;
        goto notify;
    }

    msg_page = cpu_physical_memory_map(msg_page_addr, &len, 1);
    if (!msg_page) {
        ret = -EFAULT;
        goto notify;
    }

    dst_msg = &msg_page->sint_message[sint_route->sint];

    if (dst_msg->header.message_type != HVMSG_NONE) {
        dst_msg->header.message_flags.msg_pending = 1;
        ret = -EAGAIN;
    } else {
        memcpy(dst_msg, &sint_route->msg, sizeof(*dst_msg));
        ret = kvm_hv_sint_route_set_sint(sint_route);
    }

    cpu_physical_memory_unmap(msg_page, len, 1, len);
notify:
    sint_route->msg_status = ret;
    /* notify the msg originator of the progress made; if the slot was busy we
     * set msg_pending flag in it so it will be the guest who will do EOM and
     * send the notification */
    if (ret != -EAGAIN) {
        event_notifier_set(&sint_route->sint_ack_notifier);
    }
}

/*
 * Post a Hyper-V message to the staging area, for delivery to guest in the
 * vcpu thread.
 */
void hyperv_post_msg(HvSintRoute *sint_route, struct hv_message *src_msg)
{
    assert(sint_route->msg_status == 0);

    sint_route->msg_status = -EBUSY;
    memcpy(&sint_route->msg, src_msg, sizeof(*src_msg));

    async_run_on_cpu(CPU(sint_route->cpu), cpu_post_msg, sint_route);
}

/*
 * Set given event flag for a given sint on a given vcpu, and signal the sint.
 */
int hyperv_set_evt_flag(HvSintRoute *sint_route, unsigned evtno)
{
    int ret;
    X86CPU *cpu = sint_route->cpu;
    struct hv_synic_event_flags_page *evt_page;
    unsigned long *flags;
    hwaddr evt_page_addr;
    hwaddr len = 1;
    unsigned long set_mask;
    unsigned set_idx;

    if (evtno > HV_EVENT_FLAGS_COUNT) {
        return -EINVAL;
    }

    set_idx = evtno / BITS_PER_LONG;
    set_mask = 1u << (evtno % BITS_PER_LONG);

    evt_page_addr = synic_evt_page_addr(cpu);
    if (!evt_page_addr) {
        return -ENXIO;
    }

    evt_page = cpu_physical_memory_map(evt_page_addr, &len, 1);
    if (!evt_page) {
        return -EFAULT;
    }

    flags = (unsigned long *) &evt_page->sintevent_flags[sint_route->sint];
    if ((atomic_fetch_or(&flags[set_idx], set_mask) & set_mask) != set_mask) {
        ret = kvm_hv_sint_route_set_sint(sint_route);
    } else {
        ret = 0;
    }

    cpu_physical_memory_unmap(evt_page, len, 1, len);
    return ret;
}

int kvm_hv_handle_exit(X86CPU *cpu, struct kvm_hyperv_exit *exit)
{
    CPUX86State *env = &cpu->env;

    switch (exit->type) {
    case KVM_EXIT_HYPERV_SYNIC:
        if (!cpu->hyperv_synic) {
            return -1;
        }

        switch (exit->u.synic.msr) {
        case HV_X64_MSR_SCONTROL:
            env->msr_hv_synic_control = exit->u.synic.control;
            break;
        case HV_X64_MSR_SIMP:
            env->msr_hv_synic_msg_page = exit->u.synic.msg_page;
            break;
        case HV_X64_MSR_SIEFP:
            env->msr_hv_synic_evt_page = exit->u.synic.evt_page;
            break;
        default:
            return -1;
        }
        return 0;
    case KVM_EXIT_HYPERV_HCALL: {
        uint16_t code;
        int hv_status;

        code  = exit->u.hcall.input & 0xffff;
        switch (code) {
        case HVCALL_POST_MESSAGE:
            if (!vmbus_exists()) {
                hv_status = HV_STATUS_INVALID_HYPERCALL_CODE;
                break;
            }

            hv_status = vmbus_hcall_post_message(cpu, exit);
            break;

        case HVCALL_SIGNAL_EVENT:
            if (!vmbus_exists()) {
                hv_status = HV_STATUS_INVALID_HYPERCALL_CODE;
                break;
            }

            hv_status = vmbus_hcall_signal_event(cpu, exit);
            break;

        default:
            hv_status = HV_STATUS_INVALID_HYPERCALL_CODE;
            break;
        }

        exit->u.hcall.result = hv_status;
        return 0;
    }
    default:
        return -1;
    }
}

static void kvm_hv_sint_ack_handler(EventNotifier *notifier)
{
    HvSintRoute *sint_route = container_of(notifier, HvSintRoute,
                                           sint_ack_notifier);
    event_notifier_test_and_clear(notifier);
    if (sint_route->sint_ack_clb) {
        sint_route->sint_ack_clb(sint_route);
    }
}

static void sint_route_get_sint(Object *obj, Visitor *v, const char *name,
                                void *opaque, Error **errp)
{
    uint32_t sint = SINT_ROUTE(obj)->sint;
    visit_type_uint32(v, name, &sint, errp);
}

static void sint_route_get_gsi(Object *obj, Visitor *v, const char *name,
                               void *opaque, Error **errp)
{
    int64_t gsi = SINT_ROUTE(obj)->gsi;
    visit_type_int(v, name, &gsi, errp);
}

HvSintRoute *hyperv_sint_route_new(X86CPU *cpu, uint32_t sint,
                                   const char *name,
                                   HvSintAckClb sint_ack_clb)
{
    Error *local_err = NULL;
    SynICState *synic;
    HvSintRoute *sint_route;
    int r, gsi;

    synic = SYNIC(object_resolve_path_component(OBJECT(cpu), "synic"));
    if (!synic) {
        return NULL;
    }

    sint_route = SINT_ROUTE(object_new(TYPE_SINT_ROUTE));

    r = event_notifier_init(&sint_route->sint_set_notifier, false);
    if (r) {
        goto err;
    }

    r = event_notifier_init(&sint_route->sint_ack_notifier, false);
    if (r) {
        goto err_sint_set_notifier;
    }

    event_notifier_set_handler(&sint_route->sint_ack_notifier, false,
                               kvm_hv_sint_ack_handler);

    gsi = kvm_irqchip_add_hv_sint_route(kvm_state, CPU(cpu)->cpu_index, sint);
    if (gsi < 0) {
        goto err_gsi;
    }

    r = kvm_irqchip_add_irqfd_notifier_gsi(kvm_state,
                                           &sint_route->sint_set_notifier,
                                           &sint_route->sint_ack_notifier, gsi);
    if (r) {
        goto err_irqfd;
    }

    object_property_add_child(OBJECT(synic), name, OBJECT(sint_route),
                              &local_err);
    if (local_err) {
        goto err_irqfd_notifier;
    }

    object_property_add(OBJECT(sint_route), "sint", "uint32", sint_route_get_sint,
                        NULL, NULL, NULL, NULL);
    object_property_add(OBJECT(sint_route), "gsi", "int", sint_route_get_gsi,
                        NULL, NULL, NULL, NULL);
    sint_route->gsi = gsi;
    sint_route->sint_ack_clb = sint_ack_clb;
    sint_route->cpu = cpu;
    sint_route->sint = sint;
    sint_route->refcnt = 1;
    sint_route->msg_status = 0;

    object_unref(OBJECT(sint_route));
    return sint_route;

err_irqfd_notifier:
    kvm_irqchip_remove_irqfd_notifier_gsi(kvm_state,
                                          &sint_route->sint_set_notifier, gsi);
err_irqfd:
    kvm_irqchip_release_virq(kvm_state, gsi);
err_gsi:
    event_notifier_set_handler(&sint_route->sint_ack_notifier, false, NULL);
    event_notifier_cleanup(&sint_route->sint_ack_notifier);
err_sint_set_notifier:
    event_notifier_cleanup(&sint_route->sint_set_notifier);
err:
    object_unref(OBJECT(sint_route));
    return NULL;
}

HvSintRoute *hyperv_sint_route_get(X86CPU *cpu, const char *name)
{
    HvSintRoute *sint_route;
    Object *synic = object_resolve_path_component(OBJECT(cpu), "synic");
    if (!synic) {
        return NULL;
    }

    sint_route = SINT_ROUTE(object_resolve_path_component(synic, name));
    if (!sint_route) {
        return NULL;
    }

    sint_route->refcnt++;
    return sint_route;
}

static void hyperv_sint_route_delete(HvSintRoute *sint_route)
{
    object_ref(OBJECT(sint_route));
    object_unparent(OBJECT(sint_route));
    kvm_irqchip_remove_irqfd_notifier_gsi(kvm_state,
                                          &sint_route->sint_set_notifier,
                                          sint_route->gsi);
    kvm_irqchip_release_virq(kvm_state, sint_route->gsi);
    event_notifier_set_handler(&sint_route->sint_ack_notifier, false, NULL);
    event_notifier_cleanup(&sint_route->sint_ack_notifier);
    event_notifier_cleanup(&sint_route->sint_set_notifier);
    object_unref(OBJECT(sint_route));
}

void hyperv_sint_route_unref(HvSintRoute *sint_route)
{
    if (!sint_route) {
        return;
    }
    assert(sint_route->refcnt > 0);
    if (!--sint_route->refcnt) {
        hyperv_sint_route_delete(sint_route);
    }
}

int kvm_hv_sint_route_set_sint(HvSintRoute *sint_route)
{
    return event_notifier_set(&sint_route->sint_set_notifier);
}

static void synic_get_msg_page_addr(Object *obj, Visitor *v, const char *name,
                               void *opaque, Error **errp)
{
    uint64_t value = synic_msg_page_addr(X86_CPU(obj->parent));
    visit_type_uint64(v, name, &value, errp);
}

static void synic_get_evt_page_addr(Object *obj, Visitor *v, const char *name,
                               void *opaque, Error **errp)
{
    uint64_t value = synic_evt_page_addr(X86_CPU(obj->parent));
    visit_type_uint64(v, name, &value, errp);
}

static uint64_t sint_get_msr(SINTState *state)
{
    X86CPU *cpu = X86_CPU(OBJECT(state)->parent->parent);
    return cpu->env.msr_hv_synic_sint[state->sint_no];
}

static void sint_get_vector(Object *obj, Visitor *v, const char *name,
                            void *opaque, Error **errp)
{
    uint8_t value = sint_get_msr(SINT(obj)) & HV_SYNIC_SINT_VECTOR_MASK;
    visit_type_uint8(v, name, &value, errp);
}

static void sint_get_masked(Object *obj, Visitor *v, const char *name,
                            void *opaque, Error **errp)
{
    bool value = sint_get_msr(SINT(obj)) & HV_SYNIC_SINT_MASKED;
    visit_type_bool(v, name, &value, errp);
}

static void sint_get_auto_eoi(Object *obj, Visitor *v, const char *name,
                              void *opaque, Error **errp)
{
    bool value = sint_get_msr(SINT(obj)) & HV_SYNIC_SINT_AUTO_EOI;
    visit_type_bool(v, name, &value, errp);
}

static void sint_init(Object *obj)
{
    object_property_add(obj, "vector", "uint8", sint_get_vector,
                        NULL, NULL, NULL, NULL);
    object_property_add(obj, "masked", "bool", sint_get_masked,
                        NULL, NULL, NULL, NULL);
    object_property_add(obj, "auto-eoi", "bool", sint_get_auto_eoi,
                        NULL, NULL, NULL, NULL);
}

static void synic_init(Object *obj)
{
    uint8_t i;

    object_property_add(obj, "msg-page-addr", "uint64", synic_get_msg_page_addr,
                        NULL, NULL, NULL, NULL);
    object_property_add(obj, "evt-page-addr", "uint64", synic_get_evt_page_addr,
                        NULL, NULL, NULL, NULL);

    for (i = 0; i < HV_SYNIC_SINT_COUNT; i++) {
        Object *sint_obj = object_new(TYPE_SINT);
        object_property_add_child(obj, "sint[*]", sint_obj, &error_abort);
        object_unref(sint_obj);
        SINT(sint_obj)->sint_no = i;
    }
}

void hyperv_add_synic_props(X86CPU *cpu)
{
    Object *obj;

    if (!cpu->hyperv_synic) {
        return;
    }

    obj = object_new(TYPE_SYNIC);
    object_property_add_child(OBJECT(cpu), "synic", obj, &error_abort);
    object_unref(obj);
}

static const TypeInfo synic_type_info = {
    .name = TYPE_SYNIC,
    .parent = TYPE_OBJECT,
    .instance_size = sizeof(SynICState),
    .instance_init = synic_init,
};

static const TypeInfo sint_type_info = {
    .name = TYPE_SINT,
    .parent = TYPE_OBJECT,
    .instance_size = sizeof(SINTState),
    .instance_init = sint_init,
};

static const TypeInfo sint_route_type_info = {
    .name = TYPE_SINT_ROUTE,
    .parent = TYPE_OBJECT,
    .instance_size = sizeof(HvSintRoute),
};

static void synic_register_types(void)
{
    type_register_static(&synic_type_info);
    type_register_static(&sint_type_info);
    type_register_static(&sint_route_type_info);
}

type_init(synic_register_types)
