/*
 * QEMU KVM Hyper-V support
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 *
 *
 * This header file contains additional standard hyperv definitions
 * related to hyperv and vmbus messaging protocol
 * based on drivers/hv/hyperv_vmbus.h and include/linux/hyperv.h
 *
 * These definitions are expected to move to uapi linux headers
 * and then we can start using those.
 */

#ifndef TARGET_I386_HYPERV_DEFS_H
#define TARGET_I386_HYPERV_DEFS_H

#include "standard-headers/asm-x86/hyperv.h"

typedef uint16_t hv_status;

#define HV_EVENT_FLAGS_COUNT		(256 * 8)
#define HV_EVENT_FLAGS_BYTE_COUNT	(256)
#define HV_EVENT_FLAGS_DWORD_COUNT	(256 / sizeof(uint32_t))

typedef struct {
	uint8_t b[16];
} hv_uuid;

union hv_synic_event_flags {
	uint8_t flags8[HV_EVENT_FLAGS_BYTE_COUNT];
	uint32_t flags32[HV_EVENT_FLAGS_DWORD_COUNT];
};

struct hv_synic_event_flags_page {
	union hv_synic_event_flags sintevent_flags[HV_SYNIC_SINT_COUNT];
};

union hv_connection_id {
	uint32_t asu32;
	struct {
		uint32_t id:24;
		uint32_t reserved:8;
	} u;
};

struct hv_interrupt_page
{
    unsigned long to_guest[TARGET_PAGE_SIZE / sizeof(unsigned long) / 2];
    unsigned long from_guest[TARGET_PAGE_SIZE / sizeof(unsigned long) / 2];
};

struct hv_input_post_message {
	union hv_connection_id connectionid;
	uint32_t reserved;
	uint32_t message_type;
	uint32_t payload_size;
	uint64_t payload[HV_MESSAGE_PAYLOAD_QWORD_COUNT];
};

struct hv_input_signal_event {
	union hv_connection_id connectionid;
	uint16_t flag_number;
	uint16_t rsvdz;
};

struct hv_input_signal_event_buffer {
	uint64_t align8;
	struct hv_input_signal_event event;
};

struct hv_ring_buffer {
	/* Offset in bytes from the start of ring data below */
	uint32_t write_index;

	/* Offset in bytes from the start of ring data below */
	uint32_t read_index;

	uint32_t interrupt_mask;

	/*
	 * Win8 uses some of the reserved bits to implement
	 * interrupt driven flow management. On the send side
	 * we can request that the receiver interrupt the sender
	 * when the ring transitions from being full to being able
	 * to handle a message of size "pending_send_sz".
	 *
	 * Add necessary state for this enhancement.
	 */
	uint32_t pending_send_sz;

	uint32_t reserved1[12];

	union {
		struct {
			uint32_t feat_pending_send_sz:1;
		};
		uint32_t value;
	} feature_bits;

	/* Pad it to PAGE_SIZE so that data starts on page boundary */
	uint8_t	reserved2[4028];

	/*
	 * Ring data starts here + RingDataStartOffset
	 * !!! DO NOT place any fields below this !!!
	 */
	uint8_t buffer[0];
} __attribute__((packed));

#define VMBUS_MESSAGE_TYPE 1
#define VMBUS_MESSAGE_PAYLOAD_SIZE_ALIGN sizeof(uint64_t)

/*
 * VMBUS version is 32 bit entity broken up into
 * two 16 bit quantities: major_number. minor_number.
 *
 * 0 . 13 (Windows Server 2008)
 * 1 . 1  (Windows 7)
 * 2 . 4  (Windows 8)
 * 3 . 0  (Windows 8 R2)
 * 4 . 0  (Windows 10)
 */

#define VMBUS_VERSION_WS2008  ((0 << 16) | (13))
#define VMBUS_VERSION_WIN7    ((1 << 16) | (1))
#define VMBUS_VERSION_WIN8    ((2 << 16) | (4))
#define VMBUS_VERSION_WIN8_1    ((3 << 16) | (0))
#define VMBUS_VERSION_WIN10	((4 << 16) | (0))

#define VMBUS_VERSION_INVAL -1

#define VMBUS_VERSION_CURRENT VMBUS_VERSION_WIN10

/* Make maximum size of pipe payload of 16K */
#define MAX_PIPE_DATA_PAYLOAD		(sizeof(uint8_t) * 16384)

/* Define PipeMode values. */
#define VMBUS_PIPE_TYPE_BYTE		0x00000000
#define VMBUS_PIPE_TYPE_MESSAGE		0x00000004

/* The size of the user defined data buffer for non-pipe offers. */
#define MAX_USER_DEFINED_BYTES		120

/* The size of the user defined data buffer for pipe offers. */
#define MAX_PIPE_USER_DEFINED_BYTES	116

#define VMBUS_CHANNEL_ENUMERATE_DEVICE_INTERFACE	1
#define VMBUS_CHANNEL_SERVER_SUPPORTS_TRANSFER_PAGES	2
#define VMBUS_CHANNEL_SERVER_SUPPORTS_GPADLS		4
#define VMBUS_CHANNEL_NAMED_PIPE_MODE			0x10
#define VMBUS_CHANNEL_LOOPBACK_OFFER			0x100
#define VMBUS_CHANNEL_PARENT_OFFER			0x200
#define VMBUS_CHANNEL_REQUEST_MONITORED_NOTIFICATION	0x400
#define VMBUS_CHANNEL_TLNPI_PROVIDER_OFFER		0x2000

enum vmbus_packet_type {
	VM_PKT_INVALID				= 0x0,
	VM_PKT_SYNCH				= 0x1,
	VM_PKT_ADD_XFER_PAGESET			= 0x2,
	VM_PKT_RM_XFER_PAGESET			= 0x3,
	VM_PKT_ESTABLISH_GPADL			= 0x4,
	VM_PKT_TEARDOWN_GPADL			= 0x5,
	VM_PKT_DATA_INBAND			= 0x6,
	VM_PKT_DATA_USING_XFER_PAGES		= 0x7,
	VM_PKT_DATA_USING_GPADL			= 0x8,
	VM_PKT_DATA_USING_GPA_DIRECT		= 0x9,
	VM_PKT_CANCEL_REQUEST			= 0xa,
	VM_PKT_COMP				= 0xb,
	VM_PKT_DATA_USING_ADDITIONAL_PKT	= 0xc,
	VM_PKT_ADDITIONAL_DATA			= 0xd
};

enum {
	VMBUS_MESSAGE_CONNECTION_ID	= 1,
	VMBUS_MESSAGE_PORT_ID		= 1,
	VMBUS_EVENT_CONNECTION_ID	= 2,
	VMBUS_EVENT_PORT_ID		= 2,
	VMBUS_MONITOR_CONNECTION_ID	= 3,
	VMBUS_MONITOR_PORT_ID		= 3,
	VMBUS_MESSAGE_SINT		= 2,
};

enum vmbus_channel_message_type {
	CHANNELMSG_INVALID			=  0,
	CHANNELMSG_OFFERCHANNEL		=  1,
	CHANNELMSG_RESCIND_CHANNELOFFER	=  2,
	CHANNELMSG_REQUESTOFFERS		=  3,
	CHANNELMSG_ALLOFFERS_DELIVERED	=  4,
	CHANNELMSG_OPENCHANNEL		=  5,
	CHANNELMSG_OPENCHANNEL_RESULT		=  6,
	CHANNELMSG_CLOSECHANNEL		=  7,
	CHANNELMSG_GPADL_HEADER		=  8,
	CHANNELMSG_GPADL_BODY			=  9,
	CHANNELMSG_GPADL_CREATED		= 10,
	CHANNELMSG_GPADL_TEARDOWN		= 11,
	CHANNELMSG_GPADL_TORNDOWN		= 12,
	CHANNELMSG_RELID_RELEASED		= 13,
	CHANNELMSG_INITIATE_CONTACT		= 14,
	CHANNELMSG_VERSION_RESPONSE		= 15,
	CHANNELMSG_UNLOAD			= 16,
	CHANNELMSG_UNLOAD_RESPONSE		= 17,
	CHANNELMSG_18				= 18,
	CHANNELMSG_19				= 19,
	CHANNELMSG_20				= 20,
	CHANNELMSG_TL_CONNECT_REQUEST		= 21,
	CHANNELMSG_COUNT
};

struct vmbus_channel_message_header {
	enum vmbus_channel_message_type msgtype;
	uint32_t padding;
} __attribute__((packed));

struct vmbus_channel_query_vmbus_version {
	struct vmbus_channel_message_header header;
	uint32_t version;
} __attribute__((packed));

struct vmbus_channel_initiate_contact {
	struct vmbus_channel_message_header header;
	uint32_t vmbus_version_requested;
	uint32_t target_vcpu; /* The VCPU the host should respond to */
	uint64_t interrupt_page;
	uint64_t monitor_page1;
	uint64_t monitor_page2;
} __attribute__((packed));

struct vmbus_channel_version_response {
	struct vmbus_channel_message_header header;
	uint8_t version_supported;
    uint8_t status;
} __attribute__((packed));

struct vmbus_channel_offer {
	hv_uuid if_type;
	hv_uuid if_instance;

	/*
	 * These two fields are not currently used.
	 */
	uint64_t reserved1;
	uint64_t reserved2;

	uint16_t chn_flags;
	uint16_t mmio_megabytes;		/* in bytes * 1024 * 1024 */

	union {
		/* Non-pipes: The user has MAX_USER_DEFINED_BYTES bytes. */
		struct {
			unsigned char user_def[MAX_USER_DEFINED_BYTES];
		} std;

		/*
		 * Pipes:
		 * The following sructure is an integrated pipe protocol, which
		 * is implemented on top of standard user-defined data. Pipe
		 * clients have MAX_PIPE_USER_DEFINED_BYTES left for their own
		 * use.
		 */
		struct {
			uint32_t  pipe_mode;
			unsigned char user_def[MAX_PIPE_USER_DEFINED_BYTES];
		} pipe;
	} u;
	/*
	 * The sub_channel_index is defined in win8.
	 */
	uint16_t sub_channel_index;
	uint16_t reserved3;
} __attribute__((packed));

struct vmbus_channel_offer_channel {
	struct vmbus_channel_message_header header;
	struct vmbus_channel_offer offer;
	uint32_t child_relid;
	uint8_t monitorid;
	/*
	 * win7 and beyond splits this field into a bit field.
	 */
	uint8_t monitor_allocated:1;
	uint8_t reserved:7;
	/*
	 * These are new fields added in win7 and later.
	 * Do not access these fields without checking the
	 * negotiated protocol.
	 *
	 * If "is_dedicated_interrupt" is set, we must not set the
	 * associated bit in the channel bitmap while sending the
	 * interrupt to the host.
	 *
	 * connection_id is to be used in signaling the host.
	 */
	uint16_t is_dedicated_interrupt:1;
	uint16_t reserved1:15;
	uint32_t connection_id;
} __attribute__((packed));

struct vmbus_channel_rescind_offer {
	struct vmbus_channel_message_header header;
	uint32_t child_relid;
} __attribute__((packed));

struct vmbus_gpadl_gpa_range {
    uint32_t byte_count;
    uint32_t byte_offset;
    uint64_t pfn_array[0];
};

struct vmbus_channel_gpadl_header {
	struct vmbus_channel_message_header header;
	uint32_t child_relid;
	uint32_t gpadl;
	uint16_t range_buflen;
	uint16_t rangecount;
    struct vmbus_gpadl_gpa_range range[0];
} __attribute__((packed));

/* This is the followup packet that contains more PFNs. */
struct vmbus_channel_gpadl_body {
	struct vmbus_channel_message_header header;
	uint32_t msgnumber;
	uint32_t gpadl;
	uint64_t pfn[0];
} __attribute__((packed));

struct vmbus_channel_gpadl_created {
	struct vmbus_channel_message_header header;
	uint32_t child_relid;
	uint32_t gpadl;
	uint32_t creation_status;
} __attribute__((packed));

struct vmbus_channel_gpadl_teardown {
	struct vmbus_channel_message_header header;
	uint32_t child_relid;
	uint32_t gpadl;
} __attribute__((packed));

struct vmbus_channel_gpadl_torndown {
	struct vmbus_channel_message_header header;
	uint32_t gpadl;
} __attribute__((packed));

struct vmbus_channel_open_channel {
	struct vmbus_channel_message_header header;

	/* Identifies the specific VMBus channel that is being opened. */
	uint32_t child_relid;

	/* ID making a particular open request at a channel offer unique. */
	uint32_t openid;

	/* GPADL for the channel's ring buffer. */
	uint32_t ringbuffer_gpadlhandle;

	/*
	 * Starting with win8, this field will be used to specify
	 * the target virtual processor on which to deliver the interrupt for
	 * the host to guest communication.
	 * Prior to win8, incoming channel interrupts would only
	 * be delivered on cpu 0. Setting this value to 0 would
	 * preserve the earlier behavior.
	 */
	uint32_t target_vp;

	/*
	* The upstream ring buffer begins at offset zero in the memory
	* described by RingBufferGpadlHandle. The downstream ring buffer
	* follows it at this offset (in pages).
	*/
	uint32_t downstream_ringbuffer_pageoffset;

	/* User-specific data to be passed along to the server endpoint. */
	unsigned char userdata[MAX_USER_DEFINED_BYTES];
} __attribute__((packed));

/* Open Channel Result parameters */
struct vmbus_channel_open_result {
	struct vmbus_channel_message_header header;
	uint32_t child_relid;
	uint32_t openid;
	uint32_t status;
} __attribute__((packed));

/* Close channel parameters; */
struct vmbus_channel_close_channel {
	struct vmbus_channel_message_header header;
	uint32_t child_relid;
} __attribute__((packed));

struct vmpacket_descriptor {
	uint16_t type;
	uint16_t offset8;
	uint16_t len8;
	uint16_t flags;
	uint64_t trans_id;
} __attribute__((packed));

static inline uint32_t vmpacket_get_length(struct vmpacket_descriptor *desc)
{
    assert(desc);
    return desc->len8 << 3;
}

static inline uint32_t vmpacket_get_offset(struct vmpacket_descriptor *desc)
{
    assert(desc);
    return desc->offset8 << 3;
}

struct vmpacket_mpb_array {
    uint32_t byte_count;
    uint32_t byte_offset;
    uint64_t pfn[];
};

struct vmpacket_mpb {
    struct vmpacket_descriptor desc;
    uint32_t reserved;
    uint32_t rangecount; /* always 1 */
    struct vmpacket_mpb_array range;
} __attribute__((packed));

typedef struct vmpacket_header {
    union {
        struct vmpacket_mpb mpb;
    };
} vmpacket_header;

struct vmbuspipe_hdr {
	uint32_t flags;
	uint32_t msgsize;
} __attribute__((packed));

#define ICMSGTYPE_NEGOTIATE		0
#define ICMSGTYPE_HEARTBEAT		1
#define ICMSGTYPE_KVPEXCHANGE		2
#define ICMSGTYPE_SHUTDOWN		3
#define ICMSGTYPE_TIMESYNC		4
#define ICMSGTYPE_VSS			5

#define ICMSGHDRFLAG_TRANSACTION	1
#define ICMSGHDRFLAG_REQUEST		2
#define ICMSGHDRFLAG_RESPONSE		4

struct ic_version {
	uint16_t major;
	uint16_t minor;
} __attribute__((packed));

struct icmsg_hdr {
	struct ic_version icverframe;
	uint16_t icmsgtype;
	struct ic_version icvermsg;
	uint16_t icmsgsize;
	uint32_t status;
	uint8_t ictransaction_id;
	uint8_t icflags;
	uint8_t reserved[2];
} __attribute__((packed));

struct icmsg_negotiate {
	uint16_t icframe_vercnt;
	uint16_t icmsg_vercnt;
	uint32_t reserved;
	struct ic_version icversion_data[1]; /* any size array */
} __attribute__((packed));

struct heartbeat_msg_data {
	uint64_t seq_num;
	uint32_t reserved[8];
} __attribute__((packed));

#endif
